"""some core schema (mostly lookup schema)
"""

import datajoint as dj

from loris.database.table_mixin import ManualLookup
from loris.database.attributes import lookupname


schema = dj.Schema('core')


@schema
class LookupName(ManualLookup, dj.Manual):
    primary_comment = 'identifiable name - e.g. stimulus, xml_file, array'


@schema
class ExtensionLookupName(ManualLookup, dj.Manual):
    primary_comment = 'identifiable name - e.g. prairieview, axograph'


@schema
class DataLookupName(ManualLookup, dj.Manual):
    primary_comment = 'identifiable name - e.g. stimulus, array, movie'


@schema
class FileLookupName(ManualLookup, dj.Manual):
    primary_comment = 'identifiable name - e.g. xml_file, settings'


@schema
class LookupRegex(ManualLookup, dj.Manual):
    primary_comment = 'a regular expression commonly used'



"""import all views
"""

from loris import config

# TODO import specific functions only and add __all__
from .main import *
if config['include_fly']:
    from .fly import *
from .analysis import *
from .database import *
from .erd import *
from .entries import *
from .wiki import *
from .autoscripts import *



"""class for building analysis
"""

import datetime
import os
import warnings
import pandas as pd
import numpy as np
import pickle
import json
import uuid
import glob

import datajoint as dj
from datajoint.declare import match_type
from datajoint import FreeTable
from datajoint.table import lookup_class_name
from datajoint.utils import to_camel_case
from flask import url_for
from wtforms import BooleanField, SelectField, DateField, DateTimeField, \
    StringField, FloatField, IntegerField, FormField, \
    TextAreaField, FieldList, DecimalField, HiddenField
from wtforms.validators import InputRequired, Optional, NumberRange, \
    ValidationError, Length, UUID, URL, Email
from werkzeug.datastructures import FileStorage
from wtforms import FieldList, FormField, BooleanField, StringField, \
    TextAreaField, SelectField
from wtforms.widgets import HiddenInput
from flask_wtf import FlaskForm as Form
from wtforms import Form as NoCsrfForm
from flask_wtf.file import FileField
from werkzeug.utils import secure_filename

from loris import config
from loris.utils import is_manuallookup
from loris.app.utils import name_lookup, datareader
from loris.app.forms import NONES
from loris.app.forms.formmixin import (
    FormMixin,
    ManualLookupForm, ParentFormField, DynamicFileField, DictField, ListField,
    ParentValidator, JsonSerializableValidator, AttachFileField,
    BlobFileField, Extension, TagListField, MetaHiddenField,
    ParentInputRequired, Always, LookupNameValidator, DictTextArea
)


class AutomaticFolderForm(Form, FormMixin):
    folder = DynamicFileField(
        'autoscript folder',
        description='upload your zipped autoscript folder',
        validators=[InputRequired(), Extension(['zip'])]
    )


def dynamic_scriptform(folderpath):
    class ScriptForm(NoCsrfForm, FormMixin):
        button = StringField(
            'button',
            description='button name to display',
            validators=[InputRequired()]
        )
        script = SelectField(
            'script',
            description='script to run when clicking the button',
            validators=[InputRequired()],
            choices=[
                filename
                for filename in glob.glob(os.path.join(folderpath, '*'))
                if filename.endswith('.py')
            ]
        )
        validate = TagListField(
            'validate',
            description='names of forms to validate upon pressing button',
            validators=[Optional()]
        )
        insert = BooleanField(
            'insert',
            description=(
                'whether to insert into database '
                'after running the script'
            ),
            validators=[Optional()]
        )
        configattr = StringField(
            'configattr',
            description=(
                'column name in experiment form '
                'to save the configuration file'
            ),
            validators=[Optional()]
        )
        outputfile = TagListField(
            'outputfile',
            description=(
                'name of the output file from the autoscript'
            ),
            validators=[Optional()]
        )
        outputattr = TagListField(
            'outputattr',
            description=(
                'column name in experiment form '
                'to save the output file'
            ),
            validators=[Optional()]
        )

    return ScriptForm


class FormConfig(NoCsrfForm, FormMixin):
    name = StringField(
        'name',
        description=(
            'name of the additional form'
        ),
        validators=[InputRequired()]
    )
    form_config = DictTextArea(
        'form configuration',
        description=(
            'dictionary configuration of additional form'
        ),
        validators=[InputRequired(), JsonSerializableValidator(dict)]
    )


def dynamic_formbuilder(folderpath):

    class FormBuilder(Form, FormMixin):
        buttons = FieldList(
            FormField(dynamic_scriptform(folderpath))
        )
        forms = FieldList(
            FormField(FormConfig)
        )



"""functions to create users, delete users and grant privileges
"""

import datajoint as dj

from loris import config


def change_password(username, password, connection='%'):
    """Change the password for a user. Requires administration access or
    being the user.
    """
    # establish connection
    conn = config['connection']
    # make sure all privileges have been flushed
    conn.query("FLUSH PRIVILEGES;")
    # change password
    conn.query("SET PASSWORD FOR %s@%s = %s;", (username, connection, password))
    #
    conn.query("FLUSH PRIVILEGES;")


def dropuser(username, connection='%'):
    """Drop a user from the database. Requires administration access.
    """
    # establish connection
    conn = config['connection']
    # flush privileges
    conn.query("FLUSH PRIVILEGES;")
    # drop user
    conn.query("DROP USER %s@%s;", (username, connection))
    #
    conn.query("FLUSH PRIVILEGES;")


def grantuser(
    username,
    connection='%',
    password=None,
    adduser=False,
    **privs
):
    """Add a user to the database. Requires admin/granting access.
    It also adds a user-specific schema
    """

    if password is None:
        password = config['standard_password']

    # establish connection
    conn = config['connection']

    # for safety flush all privileges
    conn.query("FLUSH PRIVILEGES;")

    # create user
    if adduser:
        conn.query(
            "DROP USER IF EXISTS %s@%s;",
            (username, connection)
        )
        conn.query(
            "CREATE USER %s@%s IDENTIFIED BY %s;",
            (username, connection, password)
        )

    # create user-specific schema
    schema = dj.Schema(username)

    privileges = {
        '*.*': "SELECT, INSERT, UPDATE, REFERENCES, CREATE",
        f'{username}.*': "ALL PRIVILEGES",
    }
    privileges.update(config['sql_privileges'])
    privileges.update(privs)

    grantprivileges(username, conn, privileges, connection)

    return schema


def grantprivileges(
    username,
    conn,
    privileges,
    connection='%',
):
    """grant privileges to user
    """

    conn.query("FLUSH PRIVILEGES;")

    for dbtable, privilege in privileges.items():
        privilege = (f"GRANT {privilege} ON {dbtable} to %s@%s;")
        conn.query(privilege, (username, connection))

    conn.query("FLUSH PRIVILEGES;")


def revokeprivileges(
    username,
    conn,
    privileges,
    connection='%',
):
    """grant privileges to user
    """

    conn.query("FLUSH PRIVILEGES;")

    for dbtable, privilege in privileges.items():
        privilege = (f"REVOKE {privilege} ON {dbtable} FROM %s@%s;")
        conn.query(privilege, (username, connection))

    conn.query("FLUSH PRIVILEGES;")



"""
    Forms
    ~~~~~
"""

from flask_wtf import FlaskForm as Form
from wtforms import (
    BooleanField, TextAreaField, StringField
)
from wtforms.validators import InputRequired, ValidationError, Optional

from loris.app.wiki.utils import clean_url
from loris.app.forms.formmixin import FormMixin
from loris.app.wiki.current_wiki import current_wiki


class URLForm(Form, FormMixin):
    url = StringField(
        'url', [InputRequired()],
        description='URL of new wiki entry'
    )

    def validate_url(form, field):
        if current_wiki.exists(field.data):
            raise ValidationError('The URL "%s" exists already.' % field.data)

    def clean_url(self, url):
        return clean_url(url)


class SearchForm(Form, FormMixin):
    term = StringField(
        'term',
        [InputRequired()],
        description='Search for... (regex accepted)'
    )
    ignore_case = BooleanField(
        'ignore case',
        default=True)


class EditorForm(Form, FormMixin):
    title = StringField(
        'title', [InputRequired()],
        description='title of wiki entry'
    )
    body = TextAreaField(
        'body', [InputRequired()],
        description='markdown text of wiki entry'
    )
    tags = StringField(
        'tags', [Optional()],
        description='tags'
    )



"""
Simple slack api
"""

import os
from slack import WebClient
from slack.errors import SlackApiError

from loris.errors import LorisError

# TODO slack blocks

def execute_slack_message(key_info, slack_info):
    """
    Parameters
    ----------
    key_info : dict
        Information of a single entry of a specific table
    slack_info : dict
        Slack information passed to slack config
    """

    token = slack_info['token']
    channel = slack_info['channel']
    text = slack_info.get('text', '').format(**key_info)

    client = WebClient(token=token)

    try:
        response = client.chat_postMessage(
            channel=channel,
            text=text
        )
        # assert response["message"]["text"] == text
    except SlackApiError as e:
        # You will get a SlackApiError if "ok" is False
        # assert e.response["ok"] is False
        # assert e.response["error"]
        # str like 'invalid_auth', 'channel_not_found'
        raise LorisError(f"Got a Slack error: {e}")

    return response



"""fly specific views
"""

import os
from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd

from loris import config
from loris.app.app import app
from loris.app.templates import form_template, joined_table_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password
from loris.io import string_load, string_dump


@app.route('/genotype', methods=['GET', 'POST'])
@login_required
def genotype():
    schema = 'subjects'
    table = 'FlyGenotype'
    subtable = None
    edit_url = url_for('genotype')
    overwrite_url = url_for('genotype')

    return form_template(
        schema, table, subtable, edit_url, overwrite_url, page='genotype',
        override_permissions=True
    )


@app.route('/stock', methods=['GET', 'POST'])
@login_required
def stock():
    schema = 'subjects'
    table = 'FlyStock'
    subtable = None
    edit_url = url_for('stock')
    overwrite_url = url_for('stock')

    return form_template(
        schema, table, subtable, edit_url, overwrite_url, page='stock',
        join_tables=[getattr(config['schemata'][schema], 'FlyGenotype')],
        joined_name='stockgenotype',
        override_permissions=True,
    )


@app.route('/cross', methods=['GET', 'POST'])
@login_required
def cross():
    schema = 'subjects'
    table = 'FlyCross'
    subtable = None
    edit_url = url_for('cross')
    overwrite_url = url_for('cross')
    load_url = url_for('crossload')

    return form_template(
        schema, table, subtable, edit_url, overwrite_url, page='cross',
        join_tables=[getattr(config['schemata'][schema], 'FlyGenotype')],
        joined_name='crossgenotype',
        load_url=load_url
    )


@app.route('/crossload', methods=['GET', 'POST'])
@login_required
def crossload():

    _id = string_load(request.args.get('_id', string_dump(None)))
    if _id is None or not isinstance(_id, dict) or 'cross_id' not in _id:
        flash('No cross_id was given for loading FlyCross', 'error')
        return redirect(url_for('cross'))

    # combine tables
    cross_table = getattr(config['schemata']['subjects'], 'FlyCross')
    genotype_table = getattr(config['schemata']['subjects'], 'FlyGenotype')

    # fetch data
    joined_table = save_join([cross_table, genotype_table])
    data = (joined_table & _id).fetch1()

    image = data['cross_schema']
    if image is not None:
        image = os.path.abspath(image)

    return render_template(
        'pages/crossload.html',
        cross_id=_id['cross_id'],
        image=image,
        experimenter=data['experimenter'],
        chromosome=f"{data['chr1']}; {data['chr2']}; {data['chr3']}",
        comments=data['comments']
    )


@app.route('/entersubject', methods=['GET', 'POST'])
@login_required
def entersubject():
    schema = 'subjects'
    table = 'FlySubject'
    subtable = None
    edit_url = url_for('entersubject')
    overwrite_url = url_for('entersubject')

    return form_template(
        schema, table, subtable, edit_url, overwrite_url, page='entersubject',
        join_tables=[getattr(config['schemata'][schema], 'FlyGenotype')],
        joined_name='subjectgenotype'
    )


@app.route('/stockgenotype', methods=['GET', 'POST'])
@login_required
def stockgenotype():
    """join various tables in the database
    """
    delete_url = url_for(
        'delete', schema='subjects', table='FlyStock', subtable=None)

    return joined_table_template(
        ['subjects.fly_genotype', 'subjects.fly_stock'],
        'Stock + Genotype Table',
        'stock',
        edit_url=url_for('stock'),
        delete_url=delete_url
    )


@app.route('/crossgenotype', methods=['GET', 'POST'])
@login_required
def crossgenotype():
    """join various tables in the database
    """
    delete_url = url_for(
        'delete', schema='subjects', table='FlyCross', subtable=None)

    return joined_table_template(
        ['subjects.fly_genotype', 'subjects.fly_cross'],
        'Cross + Genotype Table',
        'cross',
        edit_url=url_for('cross'),
        load_url=url_for('crossload'),
        delete_url=delete_url
    )


@app.route('/subjectgenotype', methods=['GET', 'POST'])
@login_required
def subjectgenotype():
    """join various tables in the database
    """
    delete_url = url_for(
        'delete', schema='subjects', table='FlySubject', subtable=None)

    return joined_table_template(
        ['subjects.fly_genotype', 'subjects.fly_subject'],
        'Subject + Genotype Table',
        'entersubject',
        edit_url=url_for('entersubject'),
        delete_url=delete_url
    )


@app.route('/stockcrossgenotype', methods=['GET', 'POST'])
@login_required
def stockcrossgenotype():
    """join various tables in the database
    """

    return joined_table_template(
        ['subjects.fly_genotype', 'subjects.fly_stock', 'subjects.fly_cross'],
        'Stock + Cross + Genotype Table',
        '#',
    )



"""factory for creating forms
"""

import pandas as pd
import numpy as np
import os
import graphviz
from flask_wtf import FlaskForm as Form
from flask import render_template, request, flash, url_for, redirect
import datajoint as dj
from datajoint.hash import key_hash
from datajoint.table import lookup_class_name
from wtforms import Form as NoCsrfForm
from wtforms import StringField, IntegerField, BooleanField, FloatField, \
    SelectField, FieldList, FormField, HiddenField

from loris.errors import LorisError
from loris import config
from loris.app.forms.dynamic_field import DynamicField
from loris.app.forms.formmixin import FormMixin, ParentFormField
from loris.app.utils import draw_helper, get_jsontable
from loris.utils import save_join


def _get_name(table):
    if hasattr(table, "name"):
        return table.name
    else:
        # split for part tables
        return dj.utils.to_camel_case(table.table_name.split('__')[-1])
    # else:
    #     return table.__name__


class DynamicForm:
    """creates forms from datajoint table class

    Parameters
    ----------
    table : class
        a subclass of datajoint.Table that is in the database.
    skip : list-like
        attributes to skip when creating form.
    formtype : class
        WTF form class or Flask form used to make a dynamic form subclass.
    """

    def __init__(self, table, skip=[], formtype=Form):

        self._table = table
        self._skip = skip
        self._formtype = formtype

        self._set()

    def reset(self):
        """reset attributes and other operations not performed in init
        """

        self._set()

    def _set(self):
        """reset attributes
        """

        self._datatable = None
        self._joined_datatable_container = {}
        self._formclass = None
        self._restriction = None
        self._fields = None
        self._part_fields = None

    @property
    def fields(self):

        if self._fields is None:
            self._fields = self.get_fields()
        return self._fields

    @property
    def part_fields(self):

        if self._part_fields is None:
            self._part_fields = self.get_part_fields()
        return self._part_fields

    @property
    def skip(self):
        return self._skip

    @property
    def formtype(self):
        return self._formtype

    @property
    def restriction(self):
        return self._restriction

    @restriction.setter
    def restriction(self, value):
        self._restriction = value

    @property
    def table(self):
        return self._table

    @property
    def joined_datatable_container(self):
        return self._joined_datatable_container

    @property
    def datatable(self):

        if self._datatable is None:
            self._datatable = self.get_datatable()

        return self._datatable

    @property
    def formclass(self):

        if self._formclass is None:
            self._formclass = self.get_formclass()

        return self._formclass

    def get_fields(self):
        """get fields attributes
        """

        fields = {}

        reserved = []
        self.table.connection.dependencies.load()
        parents = self.table.connection.dependencies.parents(self.table.full_table_name)
        for table_name, table_info in parents.items():
            if len(table_info['attr_map']) > 1:
                attr_list = []
                name_list = []
                for name in table_info['attr_map']:
                    if name in self.skip or name in reserved:
                        continue
                    # reserve and add to attribute list
                    reserved.append(name)
                    name_list.append(name)
                    attr = self.table.heading.attributes[name]
                    attr_list.append(attr)
                # empty attribute list
                if not attr_list:
                    continue
                field = DynamicField(self.table, attr_list)
                fields['___'.join(name_list)] = field

        for name, attr in self.table.heading.attributes.items():
            if name in self.skip or name in reserved:
                continue
            field = DynamicField(self.table, attr)
            fields[name] = field

        return fields

    @property
    def keys(self):
        """keys in fields
        """

        return list(self.fields.keys())

    def get_part_fields(self):
        """get part fields attribute
        """

        part_fields = {}

        for part_name in self.table.parts():
            part_table = config.get_table(part_name)
            # TODO aliased part tables
            dynamicform = type(self)(
                part_table, skip=self.table.primary_key,
                formtype=NoCsrfForm
            )
            dynamicform.restriction = self.restriction
            part_fields[_get_name(part_table)] = dynamicform

        return part_fields

    def get_formclass(self):

        class TheForm(self.formtype, FormMixin):
            pass

        for name, field in self.fields.items():
            field = field.create_field()
            if field is not None:
                setattr(
                    TheForm,
                    name,
                    field
                )

        for part_name, part_form in self.part_fields.items():
            fieldlist = FieldList(
                FormField(
                    part_form.formclass
                ),
                min_entries=1
            )
            setattr(
                TheForm,
                part_name,
                fieldlist
            )

        return TheForm

    @property
    def non_blobs(self):
        # TODO dealing with adapted attributes
        return self.table.heading.non_blobs

    def get_datatable(self):

        table = self.table

        if self.restriction is not None:
            table = table & self.restriction

        return table.proj(*self.non_blobs).fetch(
            format='frame', apply_adapter=False).reset_index()

    def get_joined_datatable(self, tables, name=None):
        """join tables with self.table and return fetched joined
        table with primary key list.
        """

        if name in self.joined_datatable_container:
            return self.joined_datatable_container[name]

        joined_table = save_join([self.table]+tables)

        if self.restriction is not None:
            joined_table = joined_table & self.restriction

        datatable = joined_table.proj(
            *joined_table.heading.non_blobs
        ).fetch(
            format='frame', apply_adapter=False
        ).reset_index()

        if name is not None:
            self.joined_datatable_container[name] = \
                datatable, joined_table.primary_key

        return datatable, joined_table.primary_key

    def get_jsontable(
        self, edit_url=None, delete_url=None, overwrite_url=None,
        join_tables=None,
        joined_name=None,
        load_url=None
    ):

        if join_tables is not None:
            table, primary_key = self.get_joined_datatable(
                join_tables, joined_name
            )
        else:
            table = self.datatable
            primary_key = self.table.primary_key

        return get_jsontable(
            table, primary_key,
            edit_url=edit_url, delete_url=delete_url,
            overwrite_url=overwrite_url, name=_get_name(self.table),
            load_url=load_url
        )

    def insert(self, form, _id=None, **kwargs):
        """insert into datajoint table

        Parameters
        ----------
        form : wtf.form from dynamicform.formclass
        _id : dict
            restriction for single entry (for save updating)
        kwargs : dict
            arguments passed to datajoint Table.insert function
        """

        if isinstance(form, FormMixin):
            formatted_dict = form.get_formatted()
        elif isinstance(form, dict):
            formatted_dict = form
        else:
            raise LorisError(f'form is incorrect type (={type(form)}); '
                             'form must be dict or be a subclass of FormMixin')

        # check if replace in kwargs affects order of main and part helper
        replace = kwargs.get('replace', False)

        def main_helper(override_update_truth=False):
            primary_dict = self._insert(
                formatted_dict, _id,
                override_update_truth=override_update_truth,
                **kwargs
            )
            return primary_dict

        def part_helper(primary_dict):
            for part_name, part_form in self.part_fields.items():
                f_dicts = formatted_dict[part_name]
                if f_dicts is None:
                    continue
                for f_dict in f_dicts:
                    if _id is None:
                        _part_id = None
                    else:
                        # update with part entry that exist
                        _part_primary = {}
                        for key, value in f_dict.items():
                            if (
                                key in part_form.table.primary_key
                                and key not in self.table.primary_key
                            ):
                                _part_primary.update(
                                    part_form.fields[key].format_value(value)
                                )
                        _part_id = {**_id, **_part_primary}
                    # insert into part table
                    part_form._insert(
                        f_dict, _part_id, primary_dict,
                        override_update_truth=True,
                        **kwargs
                    )

        if self.table.connection.in_transaction:
            if replace and _id is not None:
                part_helper(_id)
                kwargs.pop('replace')
                primary_dict = main_helper(True)
            else:
                primary_dict = main_helper()
                part_helper(primary_dict)
        else:
            # perform operations within a transaction
            with self.table.connection.transaction:
                if replace and _id is not None:
                    part_helper(_id)
                    kwargs.pop('replace')
                    primary_dict = main_helper(True)
                else:
                    primary_dict = main_helper()
                    part_helper(primary_dict)
        return primary_dict

    def _insert(
        self, formatted_dict, _id=None,
        primary_dict=None, check_reserved=True,
        override_update_truth=False,
        **kwargs
    ):
        """insert helper function
        """

        insert_dict = {}

        for key, value in formatted_dict.items():
            if key in self.fields:
                insert_dict.update(self.fields[key].format_value(value))

        if _id is None or kwargs.get('replace', False):
            truth = True
        elif len(self.table & _id) == 0:
            if override_update_truth:
                truth = True
            else:
                raise dj.DataJointError(
                    f'Entry {_id} does not exist; cannot update.'
                )
        else:
            truth = False

        if primary_dict is not None:
            insert_dict = {**primary_dict, **insert_dict}

        primary_dict = {
            key: value for key, value in insert_dict.items()
            if key in self.table.primary_key
        }

        jobs = config['schemata'][self.table.database].schema.jobs

        if check_reserved:
            reserved = (
                jobs
                & {
                    'table_name': self.table.table_name,
                    'key_hash': key_hash(primary_dict)
                }
            )
            if reserved:
                raise dj.DataJointError(
                    f"Entry {primary_dict} has been reserved for table "
                    f"{self.table.full_table_name}; "
                    "change your primary key values."
                )

        if truth:
            try:
                self.table.insert1(insert_dict, **kwargs)
            except dj.DataJointError as e:
                raise dj.DataJointError(
                    "An error occured while inserting into table "
                    f"{self.table.full_table_name}: {e}"
                )
        else:  # editing entries savely
            # DO NOT remove primary keys with new update1 method
            insert_dict = {
                key: value for key, value in insert_dict.items()
                if (
                    # key not in self.table.primary_key
                    # skip updating non-specified files
                    # TODO fix for uploading files
                    not (
                        value is None
                        and (
                            self.fields[key].attr.is_blob
                            or self.fields[key].attr.is_attachment
                        )
                    )
                )
            }
            if insert_dict and bool(set(insert_dict) - set(self.table.primary_key)):
                try:
                    self.table.update1(insert_dict)
                except dj.DataJointError as e:
                    raise dj.DataJointError(
                        "An error occured while updating table "
                        f"{self.table.full_table_name}: {e}"
                    )

        return primary_dict

    def _prepare_populate_dict(self, readonly, formatted_dict, is_edit=False):

        populate_dict = {}

        for key in self.keys:
            if key in formatted_dict:
                value = formatted_dict[key]
                if is_edit and key in self.table.primary_key:
                    readonly.append(key)
            elif '___' in key:
                value = tuple(
                    [formatted_dict[key] for key in key.split('___')]
                )
                if (
                    is_edit
                    and not (set(key.split('___'))
                             - set(self.table.primary_key))
                ):
                    readonly.append(key)
            # add to populate dict
            populate_dict[key] = self.fields[key].prepare_populate(value)

        return populate_dict

    def populate_form(
        self, restriction, form, is_edit=False, **kwargs
    ):

        readonly = []

        # restrict table
        entry = self.table & restriction

        formatted_dict = (
            entry
        ).proj(*self.non_blobs).fetch1()  # proj non_blobs?

        populate_dict = self._prepare_populate_dict(
            readonly, formatted_dict, is_edit=is_edit
        )

        # populate part tables
        for part_name in self.table.parts():
            part_table = config.get_table(part_name)
            part_formatted_list_dict = (
                part_table & restriction
            ).proj(*part_table.heading.non_blobs).fetch(as_dict=True)
            # proj non_blobs?
            populate_dict[_get_name(part_table)] = []
            for part_formatted_dict in part_formatted_list_dict:

                part_populate_dict = self.part_fields[
                    _get_name(part_table)
                ]._prepare_populate_dict(
                    readonly, part_formatted_dict, is_edit=is_edit
                )

                populate_dict[_get_name(part_table)].append(part_populate_dict)

        # update with kwargs
        populate_dict.update(kwargs)

        form.populate_form(populate_dict)
        return readonly

    def update_fields(self, form):
        """update foreign fields with new information
        """

        for field in self.fields.values():
            field.update_field(form)

    def draw_relations(self):
        """draw relations
        """

        return draw_helper(self.table)



"""Login classes
"""

from flask_login import UserMixin
import datajoint as dj
import warnings
import pymysql

from loris import config
from loris.errors import LorisError


class User(UserMixin):

    def __init__(self, user_name):
        print(config['connection'])
        self.table = config.user_table

        # skip mysql error
        try:
            if not len(self.table()):
                config.create_administrator()
        except pymysql.err.InterfaceError:
            warnings.warn('not creating administrator')

        self.user_name = user_name

        self.restriction = {config['user_name']: self.user_name}

        print(self.restriction)
        self.restricted_table = self.table & self.restriction

        if len(self.restricted_table) > 1:
            raise LorisError('More than a single user entry')

        self._entry = None

    @property
    def entry(self):

        if self._entry is None:
            if config['user_active'] is None:
                self._entry = self.restricted_table.proj(
                    config['user_name'],
                ).fetch1()

            else:
                self._entry = self.restricted_table.proj(
                    config['user_name'],
                    config['user_active'],
                ).fetch1()

        return self._entry

    @property
    def user_exists(self):
        return len(self.restricted_table)

    def get_id(self):
        return str(self.user_name)

    @property
    def is_active(self):
        if config['user_active'] is None:
            return True
        return bool(self.entry[config['user_active']])

    @property
    def is_authenticated(self):
        return self.is_active

    def check_password(self, password):
        # check password in mysql database
        try:
            config.conn(
                user=self.user_name,
                password=password,
                reset=True
            )
            success = True
        except Exception:
            success = False
        config.conn(reset=True)
        return success



"""templates
"""

from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from flask_login import current_user
import datajoint as dj
from ast import literal_eval

from loris import config
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.utils import user_has_permission, get_jsontable
from loris.utils import save_join
from loris.slack import execute_slack_message
from loris.io import string_dump, string_load


def joined_table_template(
    table_names, name='Joined Table', redirect_url='#',
    message='None', **kwargs
):
    """template for joined tables
    """

    if redirect_url != '#':
        redirect_url = url_for(redirect_url)

    tables = []
    for table_name in table_names:
        if not isinstance(table_name, str):
            tables.append(table_name)
        else:
            tables.append(config.get_table(table_name))

    joined_table = save_join(tables)
    df = joined_table.proj(
        *joined_table.heading.non_blobs
    ).fetch(format='frame').reset_index()
    data = get_jsontable(
        df, joined_table.heading.primary_key,
        **kwargs
    )

    return render_template(
        'pages/joined_table.html',
        name=name,
        url=redirect_url,
        data=data,
        toggle_off_keys=[0],
        message=message
    )


def form_template(
    schema, table, subtable, edit_url, overwrite_url, page='table',
    join_tables=None, joined_name=None, redirect_page=None,
    override_permissions=False, load_url=None, **kwargs
):
    """template for rendering tables with forms
    """

    if current_user.user_name in config['administrators']:
        readonly = []
    else:
        readonly = [config['user_name']]
        kwargs[config['user_name']] = current_user.user_name

    if redirect_page is None:
        redirect_page = page

    # showing
    enter_show = ['', 'false']
    # get id if it exists (will be restriction)
    _id = string_load(request.args.get('_id', string_dump(None)))
    # get table and create dynamic form
    table_class = getattr(config['schemata'][schema], table)
    # get table name
    table_name = '.'.join([schema, table])

    # test if user has permissions
    if _id is not None and not override_permissions:
        truth = user_has_permission(table_class & _id, current_user.user_name)
        if not truth:
            flash(
                f'User {current_user.user_name} does not have permission '
                f'to overwrite or edit entry: {_id}', 'error')
            return redirect(
                url_for(
                    redirect_page,
                    table=table,
                    schema=schema,
                    subtable=subtable,
                    _id=None
                )
            )

    # if not passed directly
    if not (subtable is None or subtable == 'None'):
        table_name = f'{table_name}.{subtable}'
        table_class = getattr(table_class, subtable)

    # standard delete and edit urls
    delete_url = url_for(
        'delete', schema=schema, table=table, subtable=subtable)

    edit = literal_eval(request.args.get('edit', "False"))
    dynamicformclass = DynamicForm

    dynamicform, form = config.get_dynamicform(
        table_name, table_class, dynamicformclass, **kwargs
    )

    filename = dynamicform.draw_relations()

    # load/notload table
    data = dynamicform.get_jsontable(
        edit_url, delete_url, overwrite_url,
        join_tables, joined_name,
        load_url=load_url
    )

    toggle_off_keys = [0]

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit is None:
            pass

        elif submit in ['Save', 'Overwrite']:
            # show enter fields
            enter_show = ['show', 'true']

            if form.validate_on_submit():
                kwargs = {}
                if submit == 'Overwrite':
                    kwargs['replace'] = True

                try:
                    _id = dynamicform.insert(
                        form,
                        (
                            _id
                            if (edit or (submit == 'Overwrite'))
                            else None
                        ),
                        **kwargs)
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    if table_name in config.slack_tables:
                        # only works with simple tables at the moment
                        if (
                            edit and
                            (
                                'edit'
                                in config.slack_tables[table_name].get(
                                    'upon', 'insert')
                                )
                        ):
                            execute_slack_message(
                                (table_class & _id).fetch1(),
                                config.slack_tables[table_name]
                            )
                        elif (
                            submit == 'Overwrite' and
                            (
                                'overwrite'
                                in config.slack_tables[table_name].get(
                                    'upon', 'insert')
                                )
                        ):
                            execute_slack_message(
                                (table_class & _id).fetch1(),
                                config.slack_tables[table_name]
                            )
                        elif 'insert' in config.slack_tables[table_name].get(
                            'upon', 'insert'
                        ):
                            execute_slack_message(
                                (table_class & _id).fetch1(),
                                config.slack_tables[table_name]
                            )
                    dynamicform.reset()
                    flash(f"Data {submit} succeeded.", 'success')
                    return redirect(
                        url_for(
                            redirect_page,
                            table=table,
                            schema=schema,
                            subtable=subtable,
                            _id=str(_id)
                        )
                    )
            else:
                flash(f"Form appears incomplete; check error messages on "
                      "fields and in pop-ups of parent tables", 'warning')

        form.append_hidden_entries()

    elif _id is not None:
        # show enter fields
        enter_show = ['show', 'true']
        try:
            readonly.extend(dynamicform.populate_form(
                _id, form, is_edit=edit,
                **kwargs
            ))
        except dj.DataJointError as e:
            flash(f"{e}", 'error')

    return render_template(
        f'pages/{page}.html',
        form=form,
        data=data,
        table_name=table_name,
        url=url_for('table', table=table, schema=schema, subtable=subtable),
        toggle_off_keys=toggle_off_keys,
        filename=filename,
        enter_show=enter_show,
        readonly=readonly
    )



"""
"""

__author__ = """gucky92"""
__email__ = 'gucky@gucky.eu'
__version__ = '0.1.4dev1'

import os

from loris.settings import Config
from loris.errors import LorisError

os.environ['DJ_SUPPORT_ADAPTED_TYPES'] = "TRUE"

loris_config_file = os.environ.get('LORIS_CONFIG_FILE', None)

config = Config.load(loris_config_file)
conn = config.conn


from loris.database.users import (
    grantuser, grantprivileges, dropuser, change_password
)


class DataBase:

    def __init__(self, config):
        self.config = config

    def __getattr__(self, name):

        if (
            ('config' in vars(self))
            and (name in self.config['schemata'])
        ):
            return self.config['schemata'][name]
        else:
            raise AttributeError(f"No database schema of name `{name}`.")


db = DataBase(config)


__all__ = [
    'config',
    'conn',
    'Config',
    'LorisError',
    'df',
    'grantuser',
    'grantprivileges',
    'dropuser',
    'change_password',
    'db'
]



"""Anatomy Tables
"""

import datajoint as dj

from loris.database.table_mixin import ManualLookup, PRIMARY_NAME, COMMENTS
from loris.database.attributes import lookupname, tags

schema = dj.Schema('anatomy')


@schema
class NeuronSection(ManualLookup, dj.Manual):
    primary_comment = 'section of a neuron - e.g. dendrite, soma'


@schema
class BrainArea(ManualLookup, dj.Manual):
    primary_comment = 'brain area - e.g. medulla'


@schema
class CellType(dj.Manual):
    definition = f"""
    {PRIMARY_NAME.format(name='cell_type', comment='standard cell type name - e.g. dm8')}
    ---
    neurotransmitters = null : <tags> # neurotransmitter of cell
    receptors = null : <tags> # receptors expressed by cell
    {COMMENTS}
    """


@schema
class ColumnId(ManualLookup, dj.Manual):
    primary_comment = 'column id - e.g. A, B, or Home'


@schema
class SourceName(ManualLookup, dj.Manual):
    primary_comment = 'source of data - e.g. fruitflybrain'


@schema
class SynapticCounts(dj.Manual):
    definition = f"""
    -> CellType.proj(postsynaptic='cell_type')  # neuron receiving input - i.e. output cell
    -> ColumnId.proj(postsynaptic_column='column_id')
    -> CellType.proj(presynaptic='cell_type')  # neuron connecting to output cell - i.e. input cell
    -> ColumnId.proj(presynaptic_column='column_id')
    ---
    synaptic_count       : int unsigned
    count_percent = null   : double  # percent in terms of postsynaptic target
    -> [nullable] SourceName
    {COMMENTS}
    """



"""dabase-speific views
"""

import os
import uuid

from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd

from loris import config
from loris.app.app import app
from loris.app.templates import form_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password
from loris.io import string_load, string_dump


@app.route('/refresh')
@login_required
def refresh():
    app.session_refresh()
    return render_template('pages/refresh.html')


@app.route('/declare', methods=['GET', 'POST'])
@login_required
def declare():
    """declare a table
    """

    form = dynamic_tablecreationform(
        current_user.user_name
    )()

    if request.method == 'POST':

        form.rm_hidden_entries()

        if form.validate_on_submit():
            try:
                form.declare_table()
                flash('Table was declared successfully', 'success')
            except dj.DataJointError as e:
                flash(f"{e}", 'error')

        form.append_hidden_entries()

    return render_template(
        'pages/declare.html',
        form=form,
        url=url_for('declare'),
    )


@app.route('/join', methods=['GET', 'POST'])
@login_required
def join():
    """join various tables in the database
    """

    formclass = dynamic_jointablesform()
    form = formclass()
    data = "None"

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit is None:
            pass

        elif submit in ['Join', 'Download']:
            if form.validate_on_submit():
                formatted_dict = form.get_formatted()

                try:
                    tables = []
                    for n, table_name in enumerate(formatted_dict['tables']):
                        tables.append(formclass.tables_dict[table_name])

                    joined_table = save_join(tables)
                    if formatted_dict['restriction'] is not None:
                        joined_table = (
                            joined_table & formatted_dict['restriction']
                        )
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    if submit == 'Download':
                        df = joined_table.fetch(format='frame').reset_index()
                        filename = str(uuid.uuid4()) + '.pkl'
                        filepath = os.path.join(
                            config['tmp_folder'], filename
                        )
                        df.to_pickle(filepath)
                        flash(
                            f"Downloaded pandas pickle file {filename} "
                            "for joined table",
                            'success'
                        )
                        return redirect(url_for('tmpfile', filename=filename))
                    else:
                        df = joined_table.proj(
                            *joined_table.heading.non_blobs
                        ).fetch(format='frame').reset_index()
                        data = get_jsontable(
                            df, joined_table.heading.primary_key
                        )
                        flash(f"Successfully joined tables.", 'success')

        form.append_hidden_entries()

    return render_template(
        'pages/join.html',
        form=form,
        url=url_for('join'),
        data=data,
        toggle_off_keys=[0]
    )


@app.route('/jobs/<schema>', methods=['GET', 'POST'])
@login_required
def jobs(schema):
    """jobs table
    """

    administrator = current_user.user_name in config['administrators']

    jobs = config['schemata'][schema].schema.jobs
    # standard delete and edit urls
    delete_url = url_for('jobs', schema=schema)
    _id = string_load(request.args.get('_id', string_dump(None)))

    data = jobs.fetch(format='frame').reset_index()
    data = get_jsontable(
        data, jobs.primary_key, delete_url=delete_url
    )

    print(administrator)

    if not administrator and _id is not None:
        flash("Only administrators are allowed to "
              "delete entries in the jobs table", "error")
    elif _id is not None:
        try:
            (jobs & _id).delete_quick()
        except dj.DataJointError as e:
            flash(f"{e}", 'error')
        else:
            flash(f'Entry deleted: {_id}', 'warning')
            return redirect(url_for('jobs', schema=schema))

    return render_template(
        'pages/jobs.html',
        data=data,
        toggle_off_keys=[0],
        schema=schema
    )

    # return form_template(
    #     schema, table, subtable,
    #     edit_url=None, overwrite_url=None,
    #     page='jobs',
    #     override_permissions=True
    # )



"""
modified from https://github.com/Viruzzz-kun/flask-pymysql/blob/master/flask_pymysql/__init__.py
"""

from flask_login import current_user

from loris import config


class MySQL(object):

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initialize the `app` for use with this
        :class:`~MySQL` class.
        This is called automatically if `app` is passed to
        :meth:`~MySQL.__init__`.

        :param flask.Flask app: the application to configure for use with
            this :class:`~flask_mysqldb.MySQL` class.
        """

        print('init sql')
        config.conn()
        self.app.config['schemata'] = list(config['schemata'].keys())
        self.app.config['tables'], self.app.config['autotables'] = \
            config.tables_to_list()
        config.close()
        print('closed connection')

        if hasattr(app, 'teardown_appcontext'):
            app.teardown_appcontext(self.teardown)

        if hasattr(app, 'before_request'):
            app.before_request(self.connect)

    def connect(self):
        print('opening')
        # TODO
        # if current_user.is_authenticated:
        #     connection = config.conn(
        #         user=current_user.user_name,
        #         password=current_user.password,
        #         reset=True
        #     )
        # else:
        connection = config.conn(
            reset=True
        )
        # not very efficient to register schema across connection
        for key, module in config['schemata'].items():
            connection.register(module.schema)
        print(connection)

    def teardown(self, exception):
        print('closing')
        config.close()



"""views for running analysis automatically
"""

import json
import os
import subprocess

from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd

from loris import config
from loris.app.app import app
from loris.app.templates import form_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform, dynamic_runform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.app.subprocess import Run
from loris.database.users import grantuser, change_password

# TODO


@app.route('/setup/<schema>/<table>', methods=['GET', 'POST'])
@login_required
def setup(schema, table):
    """setup a setting to run analysis
    """

    table_name = '.'.join([schema, table])
    url = url_for(
        'setup', schema=schema, table=table
    )

    table_class = getattr(config['schemata'][schema], table).settings_table

    delete_url = url_for(
        'delete', schema=schema, table=table_class.name
    )

    form = dynamic_settingstableform(table_class)()

    try:
        df = table_class.fetch(format='frame').reset_index()
        df = df.apply(
            lambda x: pd.Series(
                table_class()._postfetch_processing(
                    x.to_dict(), check_function=False,
                    get_joined_table=False
                )
            ),
            axis=1
        )
        df['func'] = df['func'].apply(
            lambda x: f'{x.__class__.__name__}: {x.__module__}.{x.__name__}'
        )
    except Exception as e:
        flash(str(e), 'error')
        df = table_class.proj(
            *table_class.heading.non_blobs
        ).fetch(format='frame').reset_index()

    # json table
    data = get_jsontable(
        df, table_class.primary_key,
        delete_url=delete_url
    )

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        form.rm_hidden_entries()

        if submit in ['Save', 'Overwrite']:

            if form.validate_on_submit():
                kwargs = {}
                if submit == 'Overwrite':
                    kwargs['replace'] = True

                try:
                    formatted_dict = form.get_formatted()
                    table_class.insert1(
                        formatted_dict, **kwargs
                    )
                except dj.DataJointError as e:
                    flash(f"{e}", 'error')
                else:
                    flash(f"Data {submit} succeeded.", 'success')
                    return redirect(
                        url_for(
                            'setup',
                            table=table,
                            schema=schema,
                        )
                    )

        form.append_hidden_entries()

    return render_template(
        'pages/setup.html',
        form=form,
        data=data,
        table_name=table_name,
        url=url,
        table=table,
        schema=schema,
        toggle_off_keys=[0]
    )


@app.route('/run/<schema>/<table>', methods=['GET', 'POST'])
@login_required
def run(schema, table):
    """run autopopulate on a table
    """

    table_class = getattr(config['schemata'][schema], table)
    form = dynamic_runform(table_class)()
    table_name = '.'.join([schema, table])
    # load existing process
    process = config['_autopopulate'].get(table_name, Run())

    # guidance for running new populates
    dynamicform, _form = config.get_dynamicform(
        table_name, table_class, DynamicForm
    )
    filename = dynamicform.draw_relations()
    # load/notload table
    data = dynamicform.get_jsontable()
    toggle_off_keys = [0]

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        if (
            not process.running
            and submit == 'Run'
            and form.validate_on_submit()
        ):

            formatted_dict = form.get_formatted()
            formatted_dict['reserve_jobs'] = True

            kwargs = json.dumps(formatted_dict)

            # TODO currently hard-coded
            filepath = os.path.join(
                os.path.dirname(os.path.dirname(__file__)),
                "analysis",
                "run_populate.py",
            )

            command = [
                "python",
                "-u",
                filepath,
                "--schema",
                schema,
                "--table",
                table,
                "--kwargs",
                kwargs,
                "--user",
                dj.config['database.user'],
                "--password",
                dj.config['database.password']
            ]

            config['_autopopulate'][table_name] = process
            process.start(command, config['tmp_folder'])

        elif submit == 'Abort':
            process.abort()

    stdout, stderr = process.check()

    return render_template(
        'pages/run.html',
        form=form,
        schema=schema,
        table=table,
        stdout=stdout,
        stderr=stderr,
        table_name=table_name,
        data=data,
        toggle_off_keys=toggle_off_keys,
        filename=filename
    )


@app.route('/plot/<schema>/<table>', methods=['GET', 'POST'])
@login_required
def plot(schema, table):
    flash('Plotting functionality not yet implemented', 'warning')
    return render_template('pages/home.html', user=current_user.user_name)



"""erd-specific views
"""


from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd
from ast import literal_eval

from loris import config
from loris.app.app import app
from loris.app.templates import form_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password


@app.route('/erd/', defaults={'schema': None}, methods=['GET', 'POST'])
@app.route('/erd/<schema>', methods=['GET', 'POST'])
@login_required
def erd(schema):

    only_essentials = literal_eval(request.args.get('only_essentials', 'False'))

    filename = draw_helper(
        schema, type='schema', only_essentials=only_essentials
    )

    return render_template(
        'pages/erd.html', filename=filename,
        url=url_for('erd', schema=schema),
        schema=('ERD' if schema is None else schema),
    )



"""script that is run as a subprocess when also wanting to insert
"""

import argparse
import sys
import os
import pickle

# add loris to path if not installed
try:
    from loris import config, conn
    conn()
except (ImportError):
    filepath = __file__
    for i in range(4):
        filepath = os.path.dirname(filepath)
    sys.path.append(filepath)
    from loris import config, conn
    conn()

# import other loris packages
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.subprocess import Run
from loris.errors import LorisError
from loris.app.utils import datareader, filereader
from loris.schema.core import DataLookupName, FileLookupName


def get_insert_part_mixin(
    attr, value, lookup_name, lookup_table, attr_name,
    primary_dict,
    func=lambda x: x
):
    """get dictionary to insert from part table mixin (data or file)
    """
    lookup_name = {lookup_name: attr}
    if not (lookup_table & lookup_name):
        lookup_table.insert1(lookup_name)
    return {
        **primary_dict,
        **lookup_name,
        attr_name: func(value),
    }


def data_subclass(part_table, primary_dict):

    truth = (
        set(part_table.heading)
        == (set(primary_dict) | {'a_datum', 'data_lookup_name'}))
    # TODO test attr is blob and foreign key reference

    return truth


def file_subclass(part_table, primary_dict):

    truth = (
        set(part_table.heading)
        == (set(primary_dict) | {'a_file', 'file_lookup_name'}))
    # TODO test attr is blob and foreign key reference

    return truth


def inserting_autoscript_stuff(attr, value, table_class, primary_dict):
    """inserting data/file from autoscript into database
    """
    if attr.startswith('<') and attr.endswith('>'):
        # assumes either data or filemixin was used
        attr = attr.strip('<>')
        part_table_name, attr = attr.split(':')
        part_table = getattr(table_class, part_table_name)
        if data_subclass(part_table, primary_dict):
            to_insert = get_insert_part_mixin(
                attr, value, 'data_lookup_name', DataLookupName,
                'a_datum', primary_dict, func=datareader
            )
        elif file_subclass(part_table, primary_dict):
            to_insert = get_insert_part_mixin(
                attr, value, 'file_lookup_name', FileLookupName,
                'a_file', primary_dict, func=filereader
            )
        else:
            raise LorisError(f'part table {part_table.name} is not a '
                             'subclass of DataMixin or FilesMixin.')
        part_table.insert1(to_insert)
    elif attr in table_class.heading:
        if table_class.heading[attr].is_blob:
            value = datareader(value)
        else:
            value = filereader(value)
        # update table
        table_class.update1(
            {**primary_dict, **{attr: value}}
        )
    else:
        raise LorisError(f'attr {attr} does not exist in '
                         f'table {table_class.full_table_name}')


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--location", help="location of config file", type=str)
    parser.add_argument(
        "--tablename", help="name of table in database", type=str)
    parser.add_argument(
        "--script", help="filepath to script", type=str
    )
    parser.add_argument(
        "--outputfile", help="output file after running script", type=str,
        action='append'
    )
    parser.add_argument(
        "--outputattr",
        help="attr in table for insertion of the output",
        type=str, action='append'
    )
    parser.add_argument(
        "--configattr",
        help="attr in table for insertion of configuration",
        type=str
    )

    args = parser.parse_args()
    # TODO connect as user?

    with open(args.location, 'rb') as f:
        data = pickle.load(f)

    # get table class
    schema, table = args.tablename.split('.')
    table_class = getattr(config['schemata'][schema], table)
    dynamicform = DynamicForm(table_class)

    # reserve entry
    primary_dict = {}
    for key in table_class.primary_key:
        value = data['experiment_form'][key]
        primary_dict.update(dynamicform.fields[key].format_value(value))

    print(f"Running autoscript for insertion with primary key: {primary_dict}")

    # reserve job for insertion
    jobs = config['schemata'][schema].schema.jobs
    jobs.reserve(
        table_class.table_name, primary_dict
    )

    try:
        with table_class.connection.transaction:
            # stops transaction with KeyboardInterrupt
            dynamicform.insert(
                data['experiment_form'],
                check_reserved=False
            )

            command = [
                "python",
                "-u",
                f"{args.script}",
                "--location",
                f"{args.location}",
            ]

            process = Run()
            cwd = os.path.dirname(args.script)
            process.start(command, cwd)

            lnumbers = 0
            while True:
                length = len(process.lines)
                if length > lnumbers:
                    for new_line in process.lines[lnumbers:length]:
                        print(new_line)
                    lnumbers = length
                if process.p is not None and process.p.poll() is not None:
                    break

            if process.thread.is_alive():
                process.wait()

            length = len(process.lines)
            if length > lnumbers:
                for new_line in process.lines[lnumbers:length]:
                    print(new_line)
                lnumbers = length

            # update/insert fields with data from autoscript
            # update once it starts running subprocess
            if args.configattr != 'null' and args.configattr is not None:
                inserting_autoscript_stuff(
                    args.configattr, args.location,
                    table_class, primary_dict)

            # insert output files
            # TODO option for continous update
            if args.outputattr is not None:
                for outputattr, outputfile in zip(
                    args.outputattr, args.outputfile
                ):
                    inserting_autoscript_stuff(
                        outputattr, os.path.join(cwd, outputfile),
                        table_class, primary_dict)
                    # field name or <part_table_name:data/file_lookupname>
                    # or just an attribute in the table

            if process.rc != 0:
                raise LorisError(f'automatic script error:\n{process.stderr}')

    except Exception as e:
        jobs.complete(
            table_class.table_name, primary_dict
        )
        raise e
    else:
        print(f"Finished autoscript for insertion with "
              f"primary key: {primary_dict}")
        jobs.complete(
            table_class.table_name, primary_dict
        )



"""read config file class
"""

import os
import json
import pickle
import subprocess
import uuid
import datetime
import pandas as pd
import datajoint as dj
from flask import url_for, flash
from wtforms import Form as NoCsrfForm
from flask_wtf import FlaskForm as Form
from wtforms import FormField

# use cloudpickle if installed
try:
    import cloudpickle
    pickle.dump = cloudpickle.dump
    pickle.dumps = cloudpickle.dumps
except ImportError:
    pass

from loris import config
from loris.app.utils import get_jsontable
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.formmixin import FormMixin
from loris.app.autoscripting.form_creater import dynamic_autoscripted_form
from loris.app.forms.fixed import SettingsNameForm
from loris.errors import LorisError
from loris.app.subprocess import Run


CURRENT_CONFIG = "_current_config.pkl"
SAVED_SETTINGS = "_saved_settings_{table_name}.json"
INSERT_SCRIPT = (
    f"{os.path.join(os.path.split(__file__)[0], 'run_insert_script.py')}"
)


def load_config(directory=None):
    if directory is None:
        filepath = CURRENT_CONFIG
    else:
        filepath = os.path.join(directory, CURRENT_CONFIG)
    with open(filepath, 'rb') as f:
        config = pickle.load(f)
    return config


class ConfigDynamicForm(DynamicForm):
    """changes default formtype
    """

    def __init__(self, table, skip=[], formtype=NoCsrfForm):
        super().__init__(table, skip, formtype)


class ConfigReader:
    """
    """

    def __init__(self, autoscript_filepath, table_name, **kwargs):

        if table_name is None or autoscript_filepath is None:
            self.experiment_form = "None"
            self.autoscript_forms = "None"
            self.buttons = "None"
            self.initialized = False
            self.ultra_form = "None"
            self.existing_settings = None
            self.autoscript_folder = None
            return

        self.initialized = True
        self.autoscript_folder = os.path.split(autoscript_filepath)[-1]
        self.autoscript_filepath = autoscript_filepath
        self.table_name = table_name
        self.saved_settings_file = os.path.join(
            autoscript_filepath, SAVED_SETTINGS.format(table_name=table_name))
        self.current_config_file = os.path.join(
            autoscript_filepath, CURRENT_CONFIG
        )
        if os.path.exists(self.saved_settings_file):
            self.existing_settings = pd.read_pickle(self.saved_settings_file)
        else:
            self.existing_settings = None

        schema, table = table_name.split('.')
        table_class = getattr(config['schemata'][schema], table)
        self.table_class = table_class

        # default recording id and subject id
        dynamicform, experiment_form = config.get_dynamicform(
            table_name, table_class, ConfigDynamicForm,
            **kwargs
        )
        dynamicform.reset()
        self.dynamicform = dynamicform
        self.experiment_form = experiment_form

        autoscript_forms, post_process_dict, buttons = \
            config.get_autoscriptforms(
                autoscript_filepath, table_name, dynamic_autoscripted_form,
                formclass=NoCsrfForm
            )
        self.autoscript_forms = autoscript_forms
        self.post_process_dict = post_process_dict
        self.buttons = buttons

        # dynamically create combined form
        class UltraForm(Form, FormMixin):
            table_name = self.table_class.name
            experiment_form = FormField(self.experiment_form.__class__)
            autoscript_forms = list(self.autoscript_forms.keys())
            settingsname_form = FormField(SettingsNameForm)

        for key, form in self.autoscript_forms.items():
            setattr(
                UltraForm,
                key,
                FormField(form)
            )

        self.ultra_form = UltraForm()
        # necessary currently to update experiment form in ultra form
        self.dynamicform.update_fields(self.ultra_form.experiment_form)

    def append_hidden_entries(self):

        if self.initialized:

            self.ultra_form.append_hidden_entries()

    def rm_hidden_entries(self):

        if self.initialized:

            self.ultra_form.rm_hidden_entries()

    def validate_on_submit(
        self, button=None,
        check_settings_name=False,
        flash_message=''
    ):
        """validate on submit on all forms (always check experiment form?)
        """
        truth = True

        if button is None:
            include = None
            check_experiment_form = True
        else:
            include = self.buttons[button].get('validate', [])
            if self.buttons[button].get('insert', False):
                check_experiment_form = True
            else:
                check_experiment_form = False

        if not self.ultra_form.is_submitted():
            flash(flash_message, 'error')
            return False

        if include is None:
            for key in self.ultra_form.autoscript_forms:
                truth = truth & getattr(self.ultra_form, key).form.validate()
        elif not include:
            pass
        else:
            for key in include:
                truth = truth & getattr(self.ultra_form, key).form.validate()

        if check_experiment_form:
            truth = truth & self.ultra_form.experiment_form.form.validate()

        if check_settings_name:
            truth = truth & self.ultra_form.settingsname_form.form.validate()

        if not truth:
            flash(flash_message, 'error')

        return truth

    def run(self, button):
        """run subprocess given the button key
        """

        # assert that script exists
        script = self.buttons[button].get('script')
        script_file = os.path.join(
            self.autoscript_filepath, script
        )

        assert os.path.exists(script_file), f'script {script} does not exist.'

        # get formatted form data and post_process
        keys = self.buttons[button].get('validate', [])
        data = {}
        for key in keys:
            idata = getattr(self.ultra_form, key).form.get_formatted()
            if idata is not None:
                idata = self.post_process_dict[key](idata)
            data[key] = idata

        # add experiment_form
        data['experiment_form'] = (
            self.ultra_form.experiment_form.form.get_formatted()
        )
        # save configurations to pickle file
        with open(self.current_config_file, 'wb') as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

        process = config.get('subprocess', Run())

        if not process.running:

            if self.buttons[button].get('insert', False):
                # create outputfile/attr list
                outputfiles = []
                outputattrs = []
                for outputfile, outputattr in zip(
                    self.buttons[button].get('outputfile', []),
                    self.buttons[button].get('outputattr', [])
                ):
                    outputfiles.extend(["--outputfile", f"{outputfile}"])
                    outputattrs.extend(["--outputattr", f"{outputattr}"])

                print('outputs written to:')
                print(outputfiles)
                print('outputs attributes:')
                print(outputattrs)
                # running the insert script
                command = [
                    "python",
                    "-u",
                    f"{INSERT_SCRIPT}",
                    "--tablename",
                    f"{self.table_name}",
                    "--script",
                    f"{script_file}",
                    "--location",
                    f"{self.current_config_file}",
                    "--configattr",
                    f"{self.buttons[button].get('configattr', 'null')}",
                ] + outputfiles + outputattrs
            else:
                # just run the python script
                command = [
                    "python",
                    "-u",
                    f"{script_file}",
                    "--location",
                    f"{self.current_config_file}",
                ]

            process.start(command, os.path.dirname(script_file))
            config['subprocess'] = process
            flash(f'running script {script}')
        else:
            flash(
                'Abort running subprocess before running a new process',
                'warning'
            )

    def save_settings(self):
        """save new settings to _save_settings.json - creates uuid
        for those settings
        """

        formatted_dict = self.ultra_form.get_formatted()

        settings_dict = {}
        settings_dict['_id'] = str(uuid.uuid4())
        settings_dict['name'] = formatted_dict['settingsname_form'][
            'settings_name'
        ]
        settings_dict['date'] = str(datetime.datetime.today())
        settings_dict['experiment_form'] = formatted_dict['experiment_form']

        for key in self.ultra_form.autoscript_forms:
            settings_dict[key] = formatted_dict[key]

        settings_dict = pd.Series(settings_dict)

        if self.existing_settings is None:
            self.existing_settings = pd.DataFrame([settings_dict])
        else:
            self.existing_settings = self.existing_settings.append(
                settings_dict,
                ignore_index=True,
            )

        self.existing_settings.to_pickle(self.saved_settings_file)

        flash(f'settings saved under {settings_dict["name"]}', 'success')

    def delete_settings(self, _id, name):
        """
        """

        if self.existing_settings is None:
            flash('No configurations exist', 'error')

        self.existing_settings = self.existing_settings[
            self.existing_settings['_id'] != _id
        ]

        self.existing_settings.to_pickle(self.saved_settings_file)
        flash(f'setting {name} successfully deleted', 'warning')

    def populate_form(self, _id):
        """populate form with settings table given uuid
        """

        if _id is None or not self.initialized:
            return

        selected_entries = self.existing_settings[
            self.existing_settings['_id'] == _id]

        if len(selected_entries) != 1:
            flash(f'entry id {_id} not in existing configurations.', 'error')
            return

        # settings form for diction
        settings_dict = selected_entries.iloc[0].to_dict()

        for name, field in self.dynamicform.fields.items():
            if field.is_integer:
                settings_dict['experiment_form'].pop(name, None)
            if field.is_datelike:
                settings_dict['experiment_form'].pop(name, None)
            elif field.foreign_is_manuallookup:
                fk_field = settings_dict['experiment_form'].get(name, None)
                if fk_field is not None:
                    if fk_field['existing_entries'] == '<new>':
                        fk_value = fk_field.pop(name, None)
                        settings_dict['experiment_form'][name]['existing_entries'] = str(fk_value)

        self.ultra_form.populate_form(settings_dict)

        # TODO if auto-increment and foreign_key

    def get_jsontable_settings(
        self, page='experiment', deletepage='deleteconfig'
    ):
        """get jsontable for datatables
        """

        if not self.initialized or self.existing_settings is None:
            return "None"

        return get_jsontable(
            self.existing_settings,
            primary_key=None,
            load_url=url_for(
                page,
                table_name=self.table_name,
                autoscript_folder=self.autoscript_folder,
            ),
            delete_url=url_for(
                deletepage,
                table_name=self.table_name,
                autoscript_folder=self.autoscript_folder,
            )
        )

    def get_jsontable_form(self):

        if not self.initialized:
            return "None"

        return self.dynamicform.get_jsontable(
            delete_url=url_for(
                'delete',
                schema=self.table_class.database,
                table=self.table_class.name,
            ),
            edit_url=url_for(
                'table',
                schema=self.table_class.database,
                table=self.table_class.name
            )
        )

    @property
    def toggle_off_keys(self):
        """
        """

        if self.existing_settings is None:
            return

        length = len(self.existing_settings.columns)
        # just show name and date
        return [0] + [n+3 for n in range(length-3)]



"""page class
"""

from collections import OrderedDict
from io import open
import os

from loris.app.wiki.processor import Processor


class Page(object):
    """page class
    """

    def __init__(self, path, url, new=False):
        self.path = path
        self.url = url
        self._meta = OrderedDict()
        if not new:
            self.load()
            self.render()

    def __repr__(self):
        return f"<Page: {self.url}@{self.path}>"

    def load(self):
        with open(self.path, 'r', encoding='utf-8') as f:
            self.content = f.read()

    def render(self):
        processor = Processor(self.content)
        self._html, self.body, self._meta = processor.process()

    def save(self, update=True):
        folder = os.path.dirname(self.path)
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(self.path, 'w', encoding='utf-8') as f:
            for key, value in self._meta.items():
                line = f'{key}: {value}\n'
                f.write(line)
            f.write('\n')
            f.write(self.body.replace('\r\n', '\n'))
        if update:
            self.load()
            self.render()

    @property
    def meta(self):
        return self._meta

    def __getitem__(self, name):
        return self._meta[name]

    def __setitem__(self, name, value):
        self._meta[name] = value

    @property
    def html(self):
        return self._html

    def __html__(self):
        return self.html

    @property
    def title(self):
        try:
            return self['title']
        except KeyError:
            return self.url

    @title.setter
    def title(self, value):
        self['title'] = value

    @property
    def tags(self):
        try:
            return self['tags']
        except KeyError:
            return ""

    @tags.setter
    def tags(self, value):
        self['tags'] = value



"""
Base classes and parameters that can be used to create tables
"""

COMMENTS = 'comments = null : varchar(4000)'
DESCRIPTION = 'description = null : varchar(4000) # (detailed) description'
PRIMARY_NAME = '{name} : <lookupname> # {comment}'
TAGS = 'tags = null : <tags> # comma-separated tags'
DICTTAGS = 'dict = null: <dicttags> # comma-separated tags with key=value format'
NEURAL_RECORDING = f"""
    recording_id : int auto_increment # integer id number
    ---
    recording_file_id : varchar(63) # recording file identifier -- e.g. prairieview extension
    -> subjects.FlySubject
    -> recordings.RecordingType
    -> recordings.RecordingSolution
    recording_temperature = null : float # recording temperature in Celsius
    recording_time = CURRENT_TIMESTAMP : timestamp # time of recording
    completed = 0 : <truebool> # was the recording completed as intended
    {TAGS}
    {DICTTAGS}
    {COMMENTS}
    """


class FilesMixin:
    """Part table mixin for given a master table multiple files.
    """

    @property
    def definition(self):
        return """
        -> master
        -> core.FileLookupName
        ---
        a_file : attach@attachstore
        """


class DataMixin:
    """Part table mixin for given a master table multiple files.
    """

    @property
    def definition(self):
        return """
        -> master
        -> core.DataLookupName
        ---
        a_datum : blob@datastore
        """


class ExtensionMixin:
    """Part table mixin for given a master table multiple extensions
    (not actual files).
    """

    @property
    def definition(self):
        return """
        -> master
        -> core.ExtensionLookupName
        ---
        an_extension : varchar(63)
        -> [nullable] core.LookupRegex
        """


class ManualLookup:
    """Manual table mixin with given definition
    """

    primary_comment = "short name identifier"

    @property
    def definition(self):
        return f"""
        {PRIMARY_NAME.format(name=self.table_name, comment=self.primary_comment)}
        ---
        {COMMENTS}
        """



"""
    Routes
    ~~~~~~
"""
from flask import flash, redirect, render_template, request, url_for
from flask_login import login_required

from loris.app.wiki.processor import Processor
from loris.app.forms.wiki import EditorForm, SearchForm, URLForm
from loris.app.app import app
from loris.app.wiki.current_wiki import current_wiki


@app.route('/wikihome')
@login_required
def wikihome():
    page = current_wiki.get('home')
    app.config['index_dict'] = current_wiki.index_dict()
    if page:
        return wikidisplay('home')
    return render_template('pages/wiki/home.html',
                           index_dict=app.config.get('index_dict', {}))


@app.route('/wikiindex/')
@login_required
def wikiindex():
    pages = current_wiki.index()
    return render_template(
        'pages/wiki/index.html',
        pages=pages, index_dict=app.config.get('index_dict', {}))


@app.route('/wiki/<path:url>/')
@login_required
def wikidisplay(url):
    page = current_wiki.get(url)
    if not page:
        return redirect(url_for('wikicreate', url=url))
    return render_template(
        'pages/wiki/page.html',
        page=page, index_dict=app.config.get('index_dict', {}))


@app.route('/wikicreate/', methods=['GET', 'POST'])
@login_required
def wikicreate():
    url = request.args.get('url', None)
    form = URLForm(url=url)
    if form.validate_on_submit():
        return redirect(url_for(
            'wikiedit', url=form.clean_url(form.url.data)))
    return render_template(
        'pages/wiki/create.html', form=form,
        index_dict=app.config.get('index_dict', {}))


@app.route('/wikiedit/<path:url>/', methods=['GET', 'POST'])
@login_required
def wikiedit(url):
    page = current_wiki.get(url)
    form = EditorForm(obj=page)
    if form.validate_on_submit():
        submit = request.form.get('submit', None)
        if submit == 'Cancel':
            flash(f'Wiki page "{page.title}" was not changed', 'warning')
            return redirect(url_for('wikidisplay', url=url))
        if not page:
            page = current_wiki.get_bare(url)
        form.populate_obj(page)
        page.save()
        flash(f'Wiki page "{page.title}" was saved', 'success')
        return redirect(url_for('wikidisplay', url=url))
    return render_template(
        'pages/wiki/editor.html', form=form,
        page=page, index_dict=app.config.get('index_dict', {}))


@app.route('/wikipreview/', methods=['POST'])
@login_required
def wikipreview():
    data = {}
    processor = Processor(request.form['body'])
    data['html'], data['body'], data['meta'] = processor.process()
    return data['html']


@app.route('/wikimove/<path:url>/', methods=['GET', 'POST'])
@login_required
def wikimove(url):
    page = current_wiki.get(url)
    if not page:
        return redirect(url_for('wikicreate', url=url))
    form = URLForm(obj=page)
    if form.validate_on_submit():
        newurl = form.url.data
        current_wiki.move(url, newurl)
        return redirect(url_for('wikidisplay', url=newurl))
    return render_template(
        'pages/wiki/move.html', form=form,
        page=page, index_dict=app.config.get('index_dict', {}))


@app.route('/wikidelete/<path:url>/')
@login_required
def wikidelete(url):
    page = current_wiki.get(url)
    if not page:
        return redirect(url_for('wikicreate', url=url))
    current_wiki.delete(url)
    flash(f'Wiki page "{page.title}" was deleted.', 'error')
    return redirect(url_for('wikihome'))


@app.route('/wikitags/')
@login_required
def wikitags():
    tags = current_wiki.get_tags()
    return render_template(
        'pages/wiki/tags.html',
        tags=tags, index_dict=app.config.get('index_dict', {}))


@app.route('/wikitag/<string:name>/')
@login_required
def wikitag(name):
    tagged = current_wiki.index_by_tag(name)
    return render_template(
        'pages/wiki/tag.html', pages=tagged,
        tag=name, index_dict=app.config.get('index_dict', {}))


@app.route('/wikisearch/', methods=['GET', 'POST'])
@login_required
def wikisearch():
    form = SearchForm()
    if form.validate_on_submit():
        results = current_wiki.search(form.term.data, form.ignore_case.data)
        return render_template('pages/wiki/search.html', form=form,
                               results=results, search=form.term.data,
                               index_dict=app.config.get('index_dict', {}))
    return render_template(
        'pages/wiki/search.html', form=form, search=None,
        index_dict=app.config.get('index_dict', {})
        )



"""
"""

import re

from flask import url_for


def clean_url(url):
    """
        Cleans the url and corrects various errors. Removes multiple
        spaces and all leading and trailing spaces. Changes spaces
        to underscores and makes all characters lowercase. Also
        takes care of Windows style folders use.

        :param str url: the url to clean


        :returns: the cleaned url
        :rtype: str
    """
    url = re.sub('[ ]{2,}', ' ', url).strip()
    url = url.lower().replace(' ', '_')
    url = url.replace('\\\\', '/').replace('\\', '/')
    return url


def wikilink(text, url_formatter=None):
    """
        Processes Wikilink syntax "[[Link]]" within the html body.
        This is intended to be run after content has been processed
        by markdown and is already HTML.

        :param str text: the html to highlight wiki links in.
        :param function url_formatter: which URL formatter to use,
            will by default use the flask url formatter

        Syntax:
            This accepts Wikilink syntax in the form of [[WikiLink]] or
            [[url/location|LinkName]]. Everything is referenced from the
            base location "/", therefore sub-pages need to use the
            [[page/subpage|Subpage]].

        :returns: the processed html
        :rtype: str
    """
    if url_formatter is None:
        url_formatter = url_for
    link_regex = re.compile(
        r"((?<!\<code\>)\[\[([^<].+?) \s*([|] \s* (.+?) \s*)?]])",
        re.X | re.U
    )
    for i in link_regex.findall(text):
        title = [i[-1] if i[-1] else i[1]][0]
        url = clean_url(i[1])
        html_url = u"<a href='{0}'>{1}</a>".format(
            url_formatter('wikidisplay', url=url),
            title
        )
        text = re.sub(link_regex, html_url, text, count=1)
    return text



# python ignore files and folders

ignore_file = (
    ".Python",
    "pip-log.txt",
    "pip-delete-this-directory.txt",
    ".cache",
    ".coverage",
    "local_settings.py",
    "db.sqlite3",
    "db.sqlite3-journal",
    ".webassets-cache",
    "scrapy",
    "ipython_config.py",
    "celerybeat-schedule",
    "celerybeat.pid",
    ".spyderproject",
    ".spyproject",
    ".ropeproject",
    ".dmypy.json",
    "dmypy.json"
)

ignore_file_endings = (
    ".pyc", ".pyo", ".pyd", ".py.class",
    ".so", ".manifest", ".spec",
    ".cover",
    ".pycover",
    ".mo",
    ".pot",
    ".log",
    ".ipynb_checkpoints",
    ".sage.py",
    ".env",
    ".venv"
)

ignore_file_starts = (
    ".coverage.",
)

ignore_folder = (
    "__pycache__",
    "build",
    "develop-egss",
    "dist",
    "downloads",
    "eggs",
    ".eggs",
    "lib",
    "lib64",
    "parts",
    "sdist",
    "var",
    "wheels",
    "python-wheels",
    ".installed.cfg",
    "MANIFEST",
    "htmlcov",
    ".tox",
    ".nox",
    ".coverage",
    ".cache",
    ".hypothesis",
    ".pytest_cache",
    "cover",
    "instance",
    "_build",
    ".pybuilder",
    "target",
    "profile_default",
    "__pypackages__",
    "env",
    "venv",
    "ENV",
    "env.bak",
    "venv.bak",
    "site",
    ".mypy_cache",
    ".pyre",
    ".pytype",
    "cpython_debug"
)

ignore_folder_endings = (
    ".egg-info",
    ".egg"
)

ignore_folder_starts = (
    ".coverage.",
)



"""utils functions for autoscripting
"""

from ast import literal_eval
import numpy as np
import json
import pickle
import pandas as pd


def json_reader(value):
    """read json file fields
    """
    if value is None:
        return

    with open(value, 'r') as f:
        value = json.load(f)

    return value


def array_reader(value):
    """read array file fields
    """
    if value is None:
        return

    if value.endswith('npy'):
        value = np.load(value)
    elif value.endswith('csv'):
        value = pd.read_csv(value).values
    elif value.endswith('pkl'):
        with open(value, 'rb') as f:
            value = pickle.load(f)

    assert isinstance(value, np.ndarray)
    return value


def recarray_reader(value):
    """read recarray file fields
    """
    if value is None:
        return

    return frame_reader(value).to_records(False)


def frame_reader(value):
    """read pandas dataframes file fields
    """

    if value is None:
        return

    if value.endswith('csv'):
        value = pd.read_csv(value)
    elif value.endswith('pkl'):
        with open(value, 'rb') as f:
            value = pickle.load(f)
    elif value.endswith('npy'):
        value = pd.DataFrame(np.load(value))
    elif value.endswith('json'):
        value = pd.DataFrame(json_reader(value))

    assert isinstance(value, pd.DataFrame)
    return value


def series_reader(value):
    """read pandas series file fields
    """

    return pd.Series(json_reader(value))


class ListReader:

    def __init__(self, func):

        self.func = func

    def __call__(self, value):

        if value is None:
            return

        return [self.func(val) for val in value]


class TupleReader:

    def __init__(self, func):

        self.func = func

    def __call__(self, value):

        if value is None:
            return

        return (
            None
            if val is None
            else self.func(val)
            for val in value
        )


class DictReader:

    def __init__(self, func_dict):
        self.func_dict = func_dict

    def __call__(self, value):

        if value is None:
            return

        for key, func in self.func_dict.items():
            if value[key] is None:
                pass
            else:
                value[key] = func(value[key])

        return value


class EnumReader:

    def __init__(self, value, choices):

        self.value = value
        self.choices = choices

    def __call__(self, value):

        if value is None:
            return

        index = self.choices.index(value)
        return self.value[index]


class DbReader:

    def __init__(self, table_class, columns):

        self.table_class = table_class
        self.columns = columns

    def __call__(self, value):

        if value is None:
            return

        value = literal_eval(value)
        restrict = {
            key: val
            for val, key in zip(value, self.table_class.primary_key)
        }

        entry = self.table_class & restrict
        data = entry.fetch1()

        if isinstance(self.columns, str):
            return data[self.columns]
        else:
            return [data[col] for col in self.columns]



"""
savely write and read pickle files
"""

import inspect
import cloudpickle
import pickle
import gzip
import lzma
import bz2
import os
import json
from uuid import UUID
import rapidjson

from loris import ignore
from loris.errors import LorisSerializationError


_DEFAULT_EXTENSION_MAP = {
    "gz": "gzip",
    "gzip": "gzip",
    "bz": "bz2",
    "bz2": "bz2",
    "lzma": "lzma",
    "pkl": "pickle"
}


class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return obj.hex
        return json.JSONEncoder.default(self, obj)


def string_dump(obj):
    return rapidjson.dumps(
        obj,
        uuid_mode=rapidjson.UM_CANONICAL,
        datetime_mode=rapidjson.DM_ISO8601
    )


def string_load(string):
    return rapidjson.loads(
        string.replace("'", '"'),
        uuid_mode=rapidjson.UM_CANONICAL,
        datetime_mode=rapidjson.DM_ISO8601
    )


def infer_compression_from_filename(filename: str) -> str:
    """Return the compression protocal inferred from given filename.
    Parameters
    ----------
    filename: str
        The filename for which to infer the compression protocol
    """
    return _DEFAULT_EXTENSION_MAP.get(filename.split(".")[-1], None)


def spickledumps(obj):
    """saver pickle.dumps for relative imports
    """
    # TODO doesn't work across unix and windows together
    # TODO doesn't work properly if sys path was modified!
    module = inspect.getmodule(obj)
    # if isnstance
    if module is None:
        module = inspect.getmodule(type(obj))

    # package name
    package_name = module.__name__.split('.')[0]
    # __main__ handled separately
    if package_name == '__main__':
        return b"\0" + cloudpickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
    # cwd and modulepath
    cwd = os.getcwd()
    modulepath = module.__file__
    relativepath = os.path.split(modulepath)[-1]
    # assume import of python folder from working directory
    if modulepath.startswith(cwd) and os.path.exists(package_name):
        folder_data = {}
        for root, d_names, f_names in os.walk(package_name):
            folder_name = os.path.split(root)[-1]
            if (
                folder_name in ignore.ignore_folder
                or folder_name.endswith(ignore.ignore_folder_endings)
                or folder_name.endswith(ignore.ignore_folder_starts)
            ):
                # skip unnecessary folders
                continue
            # save only allowed files
            f_names = [
                name for name in f_names
                if not (
                    name in ignore.ignore_file
                    or name.endswith(ignore.ignore_file_endings)
                    or name.startswith(ignore.ignore_file_starts)
                )
            ]
            for f_name in f_names:
                f_path = os.path.join(root, f_name)

                with open(f_path, "r") as f:
                    contents = f.read()

                folder_data[f_path] = contents
        folder_bytes = pickle.dumps(
            folder_data, protocol=pickle.HIGHEST_PROTOCOL
        )
        data_bytes = cloudpickle.dumps(
            obj, protocol=pickle.HIGHEST_PROTOCOL
        )
        return b"\1" + pickle.dumps(
            (folder_bytes, data_bytes),
            protocol=pickle.HIGHEST_PROTOCOL
        )

    # assume import of python file from working directory
    elif modulepath.startswith(cwd) and os.path.exists(relativepath):
        with open(relativepath, "r") as f:
            contents = f.read()

        folder_bytes = pickle.dumps(
            {relativepath: contents}, protocol=pickle.HIGHEST_PROTOCOL
        )
        data_bytes = cloudpickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        return b"\1" + pickle.dumps(
            (folder_bytes, data_bytes),
            protocol=pickle.HIGHEST_PROTOCOL
        )
    else:
        # just use pickledumps
        return b"\0" + cloudpickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)


def spickleloads(binary):
    """saver pickle.loads for relative imports
    """

    number = binary[0]
    data = pickle.loads(binary[1:])

    if not number:
        return data

    elif number == 1:
        folder_data = pickle.loads(data[0])
        for f_path, contents in folder_data.items():
            f_path = os.path.normpath(f_path)
            # create files in cwd if they do not exist
            if os.path.exists(f_path):
                continue

            root, f_name = os.path.split(f_path)

            # create root directories
            if not os.path.exists(root):
                os.makedirs(root)

            # write files
            with open(f_path, "w") as f:
                f.write(contents)

        return pickle.loads(data[1])

    else:
        raise LorisSerializationError(
            f"Unknown byte identifier number: '{number}'."
        )


def get_file_object(filename, mode="rb"):

    compression = infer_compression_from_filename(filename)
    if compression is None or compression == "pickle":
        file = open(filename, mode=mode)
    elif compression == "gzip":
        file = gzip.open(filename, mode=mode)
    elif compression == "bz2":
        file = bz2.open(filename, mode=mode)
    elif compression == "lzma":
        file = lzma.open(filename, mode=mode)
    else:
        raise LorisSerializationError(
            f"Compression type '{compression}' does not exist."
        )

    return file


def write_pickle(filename, data):
    """
    write pickled file
    """

    with get_file_object(filename, mode="wb") as f:
        f.write(spickledumps(data))


def read_pickle(filename):
    """
    read pickled file
    """

    with get_file_object(filename, mode="rb") as f:
        data = spickleloads(f.read())

    return data



"""subprocess class
"""

import subprocess
import threading

from flask import flash

from loris.errors import LorisError


class Run:
    """run a thread
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.p = None
        self.cmd = None
        self.cwd = None
        self.lines = []
        self.rc = None
        self.stderr = ''
        self.thread = None

    @property
    def running(self):
        return self.p is not None and self.p.poll() is None

    def start(self, cmd, cwd):
        self.reset()
        self.cmd = cmd
        self.cwd = cwd
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def run(self):

        self.p = subprocess.Popen(
            self.cmd, shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=self.cwd, bufsize=1
        )

        while True:
            output = self.p.stdout.readline()
            if self.p.poll() is not None:
                break
            if output:
                self.lines.append(output)
        self.rc = self.p.poll()
        _, self.stderr = self.p.communicate()
        # add stderr to splitlines (keeplinebreaks)
        self.lines.extend(self.stderr.splitlines(True))

    @property
    def stdout(self):
        return ''.join(self.lines)

    @property
    def lastline(self):
        if self.lines:
            return self.lines[-1]
        else:
            return ''

    def check(self):
        """check on subprocess and flash messages
        """

        if self.rc is not None:
            if self.rc == 0:
                flash('Subprocess complete', 'success')
            else:
                flash(f"Subprocess failed: "
                      f"{self.rc}", 'error')
            self.p = None
        elif self.p is not None:
            flash('Subprocess is still running', 'warning')
        else:
            flash('No subprocess is running', 'secondary')

        return self.stdout.splitlines(), self.stderr.splitlines()

    def abort(self):
        """abort subprocess
        """
        if self.p is not None:
            self.p.terminate()
            self.p = None
            flash('Aborting subprocess...', 'warning')
        else:
            flash('No subprocess is running')

    def wait(self):
        """wait for the subprocess to finish
        """

        if self.p is None:
            raise LorisError('No subprocess is running.')

        self.thread.join()

        return self.rc, self.stdout, self.stderr



"""Schema for Fly Subjects
"""

import datajoint as dj

# import schema
from loris.schema.experimenters import Experimenter
from loris.schema.anatomy import CellType

# import database mixins
from loris.database.table_mixin import (
    ManualLookup, COMMENTS, PRIMARY_NAME, TAGS
)
from loris.database.attributes import chr, link, flyidentifier, crossschema, stockid
from loris.database.attributes import lookupname, tags


schema = dj.Schema('subjects')


@schema
class FlyOrigin(ManualLookup, dj.Manual):
    primary_comment = 'where is the fly from? - e.g. Bloomington, Sarah'


@schema
class FlyGenotype(dj.Manual):
    definition = f"""
    genotype_id : int auto_increment
    ---
    date_modified : date
    chr1 : <chr>
    chr2 : <chr>
    chr3 : <chr>
    chr4 = null : <chr>
    public_ids = null : <tags>
    {TAGS}
    {COMMENTS}
    """


@schema
class FlyCross(dj.Manual):
    definition = f"""
    cross_id : int auto_increment
    ---
    date_modified : date
    -> Experimenter
    cross_schema = null : <crossschema>
    -> FlyGenotype
    status = 'planned' : enum('planned', 'crossed', 'collecting', 'terminated')
    {COMMENTS}
    """


@schema
class StockGroup(ManualLookup, dj.Manual):
    primary_comment = 'stock group name - e.g. RBF'


@schema
class FlyStock(dj.Manual):
    definition = f"""
    stock_id : <stockid>  # stock id refers to the physical stock number
    ---
    -> FlyGenotype
    -> Experimenter
    date_modified : date
    status = null : enum('dead', 'missing', 'instock', 'inpersonal', 'quarantine', 'recovery')
    priority = null : enum('1', '2', '3', 'unsure', 'drop')
    -> [nullable] FlyOrigin
    -> [nullable] FlyCross
    """


@schema
class RearingMethod(ManualLookup, dj.Manual):
    primary_comment = 'how the fly was raised - e.g. 25C, darkness'


@schema
class FlySubject(dj.Manual):
    definition = f"""
    subject_id : int auto_increment
    ---
    subject_name = null : varchar(255) # custom subject name -- not necessarily unique
    -> FlyGenotype
    -> Experimenter
    -> [nullable] RearingMethod
    sex = 'U' : enum('F', 'M', 'U')
    age = null : float # age of fly in days
    prep_time = CURRENT_TIMESTAMP : timestamp # time of prep
    {TAGS}
    {COMMENTS}
    """



"""Basic utility functions
"""


def is_manuallookup(table, ):
    """check if table is a manuallookup table
    """

    truth = False

    if (
        (len(table.heading.primary_key) == 1)
        and (len(table.heading.secondary_attributes) == 1)
    ):
        pk = table.heading.primary_key[0]
        sk = table.heading.secondary_attributes[0]

        truth = (
            (sk == 'comments')
            & (pk == table.table_name)
        )

    return truth


def save_join(tables):
    """savely join tables ignoring dependent attributes that match.
    """

    for n, table in enumerate(tables):

        if n == 0:
            joined_table = table
        else:
            dep1 = joined_table.heading.secondary_attributes
            dep2 = table.heading.secondary_attributes
            proj = list(set(dep2) - set(dep1))
            joined_table = joined_table * table.proj(*proj)

    return joined_table



"""schema for experimenters
"""


import datajoint as dj

from loris.database.table_mixin import COMMENTS, ManualLookup, PRIMARY_NAME
from loris.database.attributes import truebool
from loris.database.attributes import lookupname
from loris.database.attributes import email
from loris.database.attributes import phone


schema = dj.Schema('experimenters')


@schema
class Experimenter(dj.Manual):
    definition = f"""
    {PRIMARY_NAME.format(name='experimenter', comment='lowercase git user-name')}
    ---
    first_name : varchar(63)
    last_name : varchar(127)
    email : <email>
    phone : <phone>
    date_joined : date
    active = 0 : <truebool> # active member of the lab
    """

    class EmergencyContact(dj.Part):
        definition = f"""
        -> Experimenter
        contact_name : varchar(255)
        ---
        relation : varchar(63)
        phone : <phone>
        email = null : <email>
        {COMMENTS}
        """


@schema
class ExperimentalProject(ManualLookup, dj.Manual):
    primary_comment = 'name of experimental project'


@schema
class AssignedExperimentalProject(dj.Manual):
    definition = """
    -> Experimenter
    -> ExperimentalProject
    """



"""Class to dynamically create the different forms in the config file
"""

import os
from wtforms import (
    BooleanField, SelectField, StringField, FloatField, IntegerField,
    FormField, TextAreaField, FieldList, DecimalField
)
from wtforms.validators import InputRequired, Optional, NumberRange, \
    ValidationError, Length, UUID, URL, Email
from flask_wtf import FlaskForm as Form
from wtforms import Form as NoCsrfForm
from werkzeug.utils import secure_filename
import glob
import pandas as pd

from loris import config
from loris.app.forms import NONES
from loris.app.forms.formmixin import (
    DynamicFileField, DictField, ListField,
    JsonSerializableValidator, Extension, FormMixin
)
from loris.app.autoscripting.utils import (
    json_reader, array_reader, recarray_reader,
    frame_reader, series_reader, EnumReader, ListReader, TupleReader,
    DictReader, DbReader
)
from loris.errors import LorisError


class AutoscriptedField:

    def __init__(self, key, value, folderpath):
        self.key = key
        self.folderpath = folderpath

        if isinstance(value, (str, list)):
            self.value = value
            self.description = None
            self.default = None
            self.required = True
            self.iterate = None
            self.loc = None
        elif isinstance(value, dict):
            truth = set(value) - {'type', 'comment', 'default', 'loc', 'iterate'}
            if truth:
                raise LorisError(
                    'Key in dynamic auto-generated form contains '
                    f'illegal keywords: {truth}.'
                )
            if 'type' not in value:
                raise LorisError(
                    'Must provide type key for dynamic auto-generated form; '
                    f'only provided these keys for "{key}": {set(value)}.'
                )
            self.value = value.get('type')
            self.description = value.get('comment', None)
            self.default = value.get('default', None)
            self.required = (
                'default' not in value
                or value.get('default', None) is not None)
            self.loc = value.get('loc', None)
            self.iterate = value.get('iterate', False)
        else:
            LorisError(f"value is wrong type {type(value)}")

        self.get_field()

    def get_field(self):
        """get initialized field
        """

        self.field, self.post_process = self._get_field(
            self.key, self.value, self.required, self.default,
            self.description, self.iterate, self.loc, self.folderpath
        )

    @staticmethod
    def file_processing(value):

        if value == 'numpy.array':
            post_process = array_reader
        elif value == 'numpy.recarray':
            post_process = recarray_reader
        elif value == 'pandas.DataFrame':
            post_process = frame_reader
        elif value == 'pandas.Series':
            post_process = series_reader
        elif value == 'json':
            post_process = json_reader
        else:
            return lambda x: x

        return post_process

    @classmethod
    def _get_field(
        cls, key, value, required, default, description,
        iterate, loc, folderpath
    ):
        """get initialized field
        """

        def post_process(x):
            return x

        if required:
            kwargs = {
                'validators': [InputRequired()],
                'render_kw': {'nullable': False}
            }
        else:
            kwargs = {
                'validators': [Optional()],
                'render_kw': {'nullable': True}
            }

        kwargs['default'] = default
        kwargs['label'] = key.replace('_', ' ')
        kwargs['description'] = (key if description is None else description)

        if loc is None and not isinstance(value, dict):
            if value == 'list':
                kwargs['validators'].append(JsonSerializableValidator(list))
                field = ListField(**kwargs)
            elif value == 'dict':
                kwargs['validators'].append(JsonSerializableValidator(dict))
                field = DictField(**kwargs)
            elif value == 'str':
                field = StringField(**kwargs)
            elif value == 'set':
                kwargs['validators'].append(JsonSerializableValidator(list))
                post_process = set
                field = ListField(**kwargs)
            elif value == 'tuple':
                kwargs['validators'].append(JsonSerializableValidator(list))
                post_process = tuple
                field = ListField(**kwargs)
            elif value == 'int':
                field = IntegerField(**kwargs)
            elif value == 'float':
                field = FloatField(**kwargs)
            elif value == 'bool':
                kwargs['validators'] = [Optional()]
                field = BooleanField(**kwargs)
            elif value == 'numpy.array':
                kwargs['validators'].append(Extension())
                post_process = cls.file_processing(value)
                field = DynamicFileField(**kwargs)
            elif value == 'numpy.recarray':
                kwargs['validators'].append(Extension())
                post_process = cls.file_processing(value)
                field = DynamicFileField(**kwargs)
            elif value == 'pandas.DataFrame':
                kwargs['validators'].append(Extension())
                post_process = cls.file_processing(value)
                field = DynamicFileField(**kwargs)
            elif value == 'pandas.Series':
                kwargs['validators'].append(Extension())
                post_process = cls.file_processing(value)
                field = DynamicFileField(**kwargs)
            elif value == 'json':
                kwargs['validators'].append(Extension(['json']))
                post_process = cls.file_processing(value)
                field = DynamicFileField(**kwargs)
            elif value == 'file':
                kwargs['validators'].append(
                    Extension(config['attach_extensions']))
                field = DynamicFileField(**kwargs)
            elif isinstance(value, list):
                choices = [
                    str(ele).strip().strip('"').strip("'")
                    for ele in value
                ]
                post_process = EnumReader(value, choices)

                if default is None and not required:
                    choices = ['NULL'] + choices
                kwargs['choices'] = [(ele, ele) for ele in choices]

                field = SelectField(**kwargs)
            else:
                raise LorisError(
                    f"field value {value} not accepted for {key}."
                )
        elif loc is not None and value == 'database':
            if not isinstance(loc, list) or not len(loc) == 2:
                raise LorisError(
                    f"If type '{value}' then loc must be of type "
                    "list with exactly two elements: "
                    "1. the database table class. "
                    "2. the columns to fetch for selected entry (str or list)."
                )
            # get table from database table class name
            table = config.get_table_from_classname(loc[0])
            columns = loc[1]

            # check columns
            if isinstance(columns, str) and columns not in table.heading:
                raise LorisError(
                    f"Column '{columns}' not in table "
                    f"{table.full_table_name}; cannot create field {key}."
                )
            elif (
                not isinstance(columns, str)
                and (set(columns) - set(table.heading))
            ):
                raise LorisError(
                    f"Columns '{set(columns) - set(table.heading)}' not "
                    f"in table {table.full_table_name}; "
                    f"cannot create field {key}."
                )

            post_process = DbReader(table, columns)
            # create choices
            choices = table.proj().fetch()
            choices = [
                (str(ele), str(ele))
                if len(ele) > 1
                else (str(ele), str(ele[0]))
                for ele in choices
            ]
            choices = sorted(choices)

            if default is None and not required:
                choices = [('NULL', 'NULL')] + choices

            kwargs['choices'] = choices
            field = SelectField(**kwargs)
        elif loc is not None and isinstance(value, str):
            loc = secure_filename(loc)
            locpath = os.path.join(folderpath, loc)
            # try up to three base directories down
            if not os.path.exists(locpath):
                # try main autoscript folder
                locpath = os.path.join(os.path.dirname(folderpath), loc)
                if not os.path.exists(locpath):
                    locpath = os.path.join(
                        os.path.dirname(os.path.dirname(folderpath)), loc
                    )
                    if not os.path.exists(locpath):
                        raise LorisError(
                            f'Folder "{loc}" does not exist in '
                            f'autoscript folder '
                            f'"{os.path.basename(folderpath)}" '
                            f'and also not in the main autoscript folder.'
                        )
            # get all files from folder
            files = glob.glob(os.path.join(locpath, '*'))
            # only match certain extensions
            if (value == 'pandas.DataFrame') or (value == 'numpy.recarray'):
                files = [
                    ifile for ifile in files
                    if (
                        ifile.endswith('.pkl')
                        or ifile.endswith('.npy')
                        or ifile.endswith('.csv')
                        or ifile.endswith('.json')
                    )
                ]
            elif value == 'numpy.array':
                files = [
                    ifile for ifile in files
                    if (
                        ifile.endswith('.pkl')
                        or ifile.endswith('.npy')
                        or ifile.endswith('.csv')
                    )
                ]
            elif (value == 'json') or (value == 'pandas.Series'):
                files = [ifile for ifile in files if ifile.endswith('.json')]
            else:
                # skip file that start with two underscores e.g. __init__.py
                files = [
                    ifile
                    for ifile in files
                    if not os.path.basename(ifile).startswith('__')
                ]
            # setup as choices
            choices = [
                (str(ele), os.path.split(ele)[-1])
                for ele in files
            ]
            # setup None choice
            if default is None and not required:
                choices = [('NULL', 'NULL')] + choices
            kwargs['choices'] = choices
            post_process = cls.file_processing(value)
            field = SelectField(**kwargs)
        elif isinstance(value, dict):
            form, post_process = dynamic_autoscripted_form(
                value, folderpath, NoCsrfForm
            )
            field = FormField(form)
        # TODO set number of fieldlists (startswith numeric)
        else:
            raise LorisError(f"field value {value} not accepted for {key}.")

        # make iterator (can add multiple values together)
        if iterate:
            field = FieldList(
                field,
                min_entries=int(required) + 1  # have one required field if so
            )
            post_process = ListReader(post_process)
        return field, post_process


def dynamic_autoscripted_form(dictionary, folderpath, formclass=Form):

    post_process_dict = {}

    class DynamicForm(formclass, FormMixin):
        pass

    for key, value in dictionary.items():
        # comments in the json or formatting guidelines start with _
        if key.startswith('#'):
            continue
        if not key.isidentifier():
            raise LorisError(
                f"key {key} is not an identifier; i.e. alphanumeric "
                "and underscore characters only. The key needs to be "
                "an identifier if it is used as a keyword during function "
                "calling."
            )

        auto_field = AutoscriptedField(
            key, value, folderpath
        )

        post_process_dict[key] = auto_field.post_process

        setattr(
            DynamicForm,
            key,
            auto_field.field
        )

    return DynamicForm, DictReader(post_process_dict)



"""
"""

from flask import Flask, request, redirect
from flask_login import LoginManager

from loris import config
from loris.app.login import User
from loris.app.pymysql_flask import MySQL


class LorisApp(Flask):

    def session_refresh(self):
        config.refresh()
        # for testing when refresh happens
        self.config['schemata'] = list(config['schemata'].keys())
        self.config['tables'], self.config['autotables'] = \
            config.tables_to_list()


app = LorisApp(__name__)
app.secret_key = config['secret_key']
app.config['include_fly'] = config['include_fly']
app.config['external_wiki'] = config['external_wiki']

mysql = MySQL(app)
login_manager = LoginManager(app)


@login_manager.user_loader
def load_user(user_id):
    return User(user_id)


@login_manager.unauthorized_handler
def unauthorized_callback():
    return redirect('/login?target=' + request.path)


from loris.app import views, errors



"""
"""

from flask import render_template, request, flash, url_for, redirect, \
    send_from_directory, session, Response
from functools import wraps
from flask_login import current_user, login_user, login_required, logout_user
import datajoint as dj
import pandas as pd
import json
import os
import time

from loris import config
from loris.app.app import app
from loris.app.templates import form_template
from loris.app.forms.dynamic_form import DynamicForm
from loris.app.forms.fixed import (
    dynamic_jointablesform, dynamic_settingstableform, LoginForm,
    PasswordForm, dynamic_tablecreationform, dynamic_autoscriptform,
)
from loris.app.utils import draw_helper, get_jsontable, user_has_permission
from loris.utils import save_join
from loris.app.login import User
from loris.database.users import grantuser, change_password
from loris.app.autoscripting.config_reader import ConfigReader
from loris.app.subprocess import Run


# TODO work on monitor streaming


def stream_template(template_name, **context):
    app.update_template_context(context)
    t = app.jinja_env.get_template(template_name)
    rv = t.stream(context)
    rv.enable_buffering(5)
    return rv


def gen_stdout(process):
    lines_inserted = []
    while process.running:
        new_lines = [
            line
            for line in process.lines
            if line not in lines_inserted
        ]
        if new_lines:
            lines_inserted.extend(new_lines)
            # new added lines need to be broken too
            yield '<br>'.join(new_lines) + "<br>"
    new_lines = [
        line
        for line in process.lines
        if line not in lines_inserted
    ]
    lines_inserted.extend(new_lines)
    if new_lines:
        yield '<br>'.join(new_lines)


@app.route('/generate_stdout')
def generate_stdout():
    return Response(
        gen_stdout(config.get('subprocess', Run())),
        mimetype='multipart/x-mixed-replace'
    )


@app.route("/monitorexperiment")
@login_required
def monitorexperiment():
    return render_template(
        'pages/monitorexperiment.html',
    )
    # return Response(generate_stdout())
    stdout, stderr = config.get('subprocess', Run()).check()
    return Response(
        stream_template(
            'pages/monitorexperiment.html',
            stdout=stdout, stderr=stderr
        )
    )


@app.route("/buildexperiment")
@login_required
def buildexperiment():
    flash('Build Experiment functionality not yet implemented', 'warning')
    return render_template('pages/home.html', user=current_user.user_name)


@app.route("/experiment",
           defaults={'table_name': None, 'autoscript_folder': None},
           methods=['GET', 'POST'])
@app.route("/experiment/<table_name>/<autoscript_folder>",
           methods=['GET', 'POST'])
@login_required
def experiment(table_name, autoscript_folder):

    _id = request.args.get('_id', None)
    folder = config['autoscript_folder']

    if autoscript_folder is not None and not autoscript_folder == 'None':
        autoscript_filepath = os.path.join(folder, autoscript_folder)
    else:
        autoscript_filepath = None

    reader = ConfigReader(autoscript_filepath, table_name,)
    if reader.initialized:
        enter_show = ['show', 'true']
    else:
        enter_show = ['', 'false']

    form = dynamic_autoscriptform()(prefix='load')

    submit = request.args.get('submit', None)

    process = config.get('subprocess', Run())

    if request.method == 'POST':

        reader.rm_hidden_entries()

        submit = request.form.get('submit', None)
        print(submit)

        if submit == 'Load' and form.validate_on_submit():
            autoscript_filepath = form.autoscript.data
            table_name = form.experiment.data

            return redirect(url_for(
                'experiment',
                autoscript_folder=os.path.basename(autoscript_filepath),
                table_name=table_name))

        elif submit == 'Abort':
            process.abort()

        elif reader.initialized:
            if (
                (submit in reader.buttons)
                and reader.validate_on_submit(
                    submit,
                    flash_message=(
                        'Unable to run process; '
                        'check all fields in the forms')
                )
            ):
                reader.run(submit)
                # return redirect(url_for('monitorexperiment'))

            elif (
                (submit == 'Save')
                and reader.validate_on_submit(
                    check_settings_name=True,
                    flash_message=(
                        'Unable to save settings; '
                        'check all fields in the forms')
                )
            ):
                reader.save_settings()

        # append hidden entries
        reader.append_hidden_entries()

    else:
        reader.populate_form(_id)

    stdout, stderr = process.check()

    return render_template(
        f'pages/experiment.html',
        form=form,
        data=reader.get_jsontable_settings(),
        data2=reader.get_jsontable_form(),
        toggle_off_keys2=[0],
        toggle_off_keys=reader.toggle_off_keys,
        ultra_form=reader.ultra_form,
        buttons=reader.buttons,
        enter_show=enter_show,
        stdout=stdout,
        stderr=stderr,
        url_experiment=(None if table_name is None else url_for(
            'table',
            **{
                key: value
                for key, value in
                zip(('schema', 'table', 'subtable'), table_name.split('.'))
            }
        ))
    )


@app.route("/deleteconfig/<table_name>/<autoscript_folder>",
           methods=['GET', 'POST'])
@login_required
def deleteconfig(table_name, autoscript_folder):

    redirect_url = request.args.get(
        'target',
        url_for(
            'experiment',
            table_name=table_name,
            autoscript_folder=autoscript_folder)
    )

    _id = request.args.get('_id', None)
    autoscript_filepath = os.path.join(
        config['autoscript_folder'], autoscript_folder
    )

    if _id is None:
        return redirect(redirect_url)

    reader = ConfigReader(autoscript_filepath, table_name,)

    name = reader.existing_settings[
        reader.existing_settings['_id'] == _id
    ]['name']

    if len(name) != 1:
        return redirect(redirect_url)

    name = name.iloc[0]

    if request.method == 'POST':
        submit = request.form.get('submit', None)

        if submit == 'Delete':
            reader.delete_settings(_id, name)
        else:
            flash(f'Entry {name} was not deleted')
        return redirect(redirect_url)

    message = (
        f"Are you sure you want to delete configuration <<{name}>> "
        f"for protocol <<{autoscript_folder}>> and experiment type <<{table_name}>>?")

    return render_template(
        'pages/deleteconfig.html',
        message=message
    )



"""
"""

import graphviz
import os
import pandas as pd
import uuid
import pickle
import json
import glob
import shutil
import datajoint as dj
import numpy as np
from datajoint.schemas import lookup_class_name
from flask import url_for

from loris import config
from loris.errors import LorisError
from loris.utils import is_manuallookup
from loris.io import string_dump


# TODO deprecated
user_has_permission = config.user_has_permission


def datareader(value):
    """data reader for supported files, otherwise just return the file
    """
    if value.endswith('npy') or value.endswith('npz'):
        value = np.load(value)
    elif value.endswith('csv'):
        value = pd.read_csv(value).to_records(False)
    elif value.endswith('pkl'):
        with open(value, 'rb') as f:
            value = pickle.load(f)
    elif value.endswith('json'):
        with open(value, 'r') as f:
            value = json.load(f)

    return value


def filereader(value):
    """
    Copies filepath to tmp folder with uuid
    """
    if not os.path.exists(value):
        raise LorisError(f"Filepath {value} does not exist.")
    basename = os.path.basename(value)
    basename = f"{uuid.uuid4()}_{basename}"
    dst = os.path.join(config['tmp_folder'], basename)
    assert not os.path.exists(dst)
    shutil.copyfile(value, dst)
    return dst


def get_jsontable(
    data, primary_key, edit_url=None, delete_url=None,
    overwrite_url=None, name=None, load_url=None
):
    """get json table from dataframe
    """

    if primary_key is None:
        pass
    elif len(data) == 0:
        data = pd.DataFrame(
            columns=['_id'] + list(data.columns)
        )
    else:
        _id = pd.Series(
            data[primary_key].to_dict('records')
        ).apply(string_dump)
        _id.name = '_id'
        data = pd.concat([_id, data], axis=1)

    jsontable = {}
    jsontable['delete_url'] = str(delete_url)
    jsontable['edit_url'] = str(edit_url)
    jsontable['load_url'] = str(load_url)
    jsontable['overwrite_url'] = str(overwrite_url)
    jsontable['execute'] = 'True'
    jsontable['id'] = str(name)
    jsontable['head'] = list(data.columns)
    jsontable['data'] = data.values
    return jsontable


def name_lookup(full_name):
    """ Look for a table's class name given its full name. """
    try:
        return config.get_table(full_name, as_string=True)
    except Exception:
        return full_name


def draw_helper(obj=None, type='table', only_essentials=False):
    """
    helper for drawing erds
    """
    # Do not redo when image exists
    if obj is None:
        filename = 'eerrdd'
    elif type == 'table':
        filename = obj.full_table_name
    else:
        filename = obj

    filename = filename + str(only_essentials)

    filepath = os.path.join(config['tmp_folder'], filename)

    # append filename to empty list for removing when refreshing database
    config['_empty'].append(filename)

    filepaths = glob.glob(filepath + '*' + '.svg')
    if len(filepaths) == 1:
        return os.path.split(filepaths[0])[-1]

    # add random string (for rendering purposes on browsers)
    random_string = str(uuid.uuid4())
    filename += random_string
    filepath += random_string

    # rankdir TB?
    # setup of graphviz
    graph_attr = {
        'rankdir': 'LR', 'splines': 'ortho',
        'fontname': 'helvetica'
    }
    node_attr = {
        'style': 'filled, rounded', 'shape': 'record', 'align': 'center',
        'ranksep': '0.1', 'fontsize': '10',
        'penwidth': '1.5', 'height': '0.1', 'fontname': 'helvetica'
    }
    edge_attr = {
        'arrowsize': '0.7',
        'headclip': 'true',
        'penwidth': '1.5'
    }

    node_attrs = {
        dj.Manual: {
            'fillcolor': 'darkgreen',
            'color': 'darkgreen',
            'fontsize': '10',
            'fontcolor': 'white'},
        dj.Computed: {
            'fillcolor': 'darkorchid3',
            'color': 'darkorchid3',
            'fontsize': '10',
            'fontcolor': 'white'},
        dj.Lookup: {
            'fillcolor': 'azure4',
            'color': 'azure4',
            'fontsize': '10',
            'fontcolor': 'white'},
        dj.Imported: {
            'fillcolor': 'navyblue',
            'color': 'navyblue',
            'fontsize': '10',
            'fontcolor': 'white'},
        dj.Part: {
            'fillcolor': 'azure4',
            'color': 'azure4',
            'fontsize': '6',
            'fontcolor': 'white'},
        dj.Settingstable: {
            'fillcolor': 'goldenrod',
            'color': 'goldenrod',
            'fontsize': '6',
            'fontcolor': 'white'},
        dj.AutoComputed: {
            'fillcolor': 'darkorchid4',
            'color': 'darkorchid4',
            'fontsize': '10',
            'fontcolor': 'white'},
        dj.AutoImported: {
            'fillcolor': 'navy',
            'color': 'navy',
            'fontsize': '10',
            'fontcolor': 'white'},
    }

    edge_attrs = {
        dj.Manual: {
            'color': 'darkgreen',
        },
        dj.Computed: {
            'color': 'darkorchid3',
        },
        dj.Lookup: {
            'color': 'azure4',
        },
        dj.Imported: {
            'color': 'navyblue',
        },
        dj.Part: {
            'color': 'azure4',
        },
        dj.Settingstable: {
            'color': 'goldenrod',
        },
        dj.AutoComputed: {
            'color': 'darkorchid4',
        },
        dj.AutoImported: {
            'color': 'navy',
        },
    }

    dot = graphviz.Digraph(
        graph_attr=graph_attr, node_attr=node_attr,
        edge_attr=edge_attr, engine='dot', format='svg')

    def add_node(name, node_attr={}):
        """
        Add a node/table to the current graph (adding subgraphs if needed).
        """

        table_names = dict(
            zip(['schema', 'table', 'subtable'], name.split('.'))
        )
        graph_attr = {
            'color': 'black',
            'label': '<<B>{}</B>>'.format(table_names['schema']),
            'URL': url_for('erd', schema=table_names['schema']),
            'target': '_top',
            'style': 'rounded',
            'penwidth': '3'
        }

        with dot.subgraph(
            name='cluster_{}'.format(table_names['schema']),
            node_attr=node_attr,
            graph_attr=graph_attr
        ) as subgraph:
            subgraph.node(
                name,
                label='{}'.format(name.split('.')[-1]),
                URL=url_for('table', **table_names),
                target='_top', **node_attr
            )
        return name

    def is_essential(name):
        truth = dj.diagram._get_tier(name) in [
            dj.Manual, dj.Computed, dj.Lookup,
            dj.Imported, dj.AutoComputed, dj.AutoImported
        ]

        if truth:
            try:
                schema, table = name_lookup(name).split('.')
                table = getattr(
                    config['schemata'][schema],
                    table
                )
                truth = not (
                    is_manuallookup(table)
                    or (
                        table.full_table_name
                        == config.user_table.full_table_name)
                    or (
                        table.full_table_name
                        == config.group_table.full_table_name)
                    or (
                        table.full_table_name
                        == config.assigned_table.full_table_name)
                    or (
                        schema
                        in config.group_table.proj().fetch()[
                            config['group_name']
                        ]
                    ) or (
                        schema
                        in config.user_table.proj().fetch()[
                            config['user_name']
                        ]
                    )
                )
            except (KeyError, ValueError):
                print('did not check essential table')

        return truth

    if type == 'table':
        root_table = obj
        root_dependencies = root_table.connection.dependencies
        root_dependencies.load()
        root_name = root_table.full_table_name
        root_id = add_node(
            name_lookup(root_name), node_attrs[dj.diagram._get_tier(root_name)]
        )

        # in edges
        for node_name, _ in root_dependencies.in_edges(root_name):
            if dj.diagram._get_tier(node_name) is dj.diagram._AliasNode:
                # renamed attribute
                node_name = list(root_dependencies.in_edges(node_name))[0][0]

            node_id = add_node(
                name_lookup(node_name),
                node_attrs[dj.diagram._get_tier(node_name)]
            )
            dot.edge(
                node_id, root_id,
                **edge_attrs[dj.diagram._get_tier(node_name)]
            )

        # out edges
        for _, node_name in root_dependencies.out_edges(root_name):
            if dj.diagram._get_tier(node_name) is dj.diagram._AliasNode:
                # renamed attribute
                node_name = list(root_dependencies.out_edges(node_name))[0][1]

            node_id = add_node(
                name_lookup(node_name),
                node_attrs[dj.diagram._get_tier(node_name)]
            )
            dot.edge(
                root_id, node_id,
                **edge_attrs[dj.diagram._get_tier(root_name)]
            )
    else:
        dependencies = config['connection'].dependencies
        dependencies.load()
        for root_name in dependencies.nodes.keys():
            try:
                int(root_name)
                continue
            except Exception:
                pass
            schema = root_name.replace('`', '').split('.')[0]
            if obj is None and schema in config['skip_schemas']:
                continue
            if obj is not None and (obj != schema):
                continue
            if only_essentials and not is_essential(root_name):
                continue

            root_id = add_node(
                name_lookup(root_name),
                node_attrs[dj.diagram._get_tier(root_name)]
            )

            for _, node_name in dependencies.out_edges(root_name):
                if dj.diagram._get_tier(node_name) is dj.diagram._AliasNode:
                    # renamed attribute
                    node_name = list(dependencies.out_edges(node_name))[0][1]
                if only_essentials and not is_essential(node_name):
                    continue

                node_id = add_node(
                    name_lookup(node_name),
                    node_attrs[dj.diagram._get_tier(node_name)]
                )
                dot.edge(
                    root_id, node_id,
                    **edge_attrs[dj.diagram._get_tier(root_name)]
                )

    if os.path.exists(filepath):
        os.remove(filepath)
    dot.render(filepath)

    return f'{filename}.svg'



"""class for configurations
"""

import json
import os
import sys
import shutil
from getpass import getpass
import importlib
import inspect
import multiprocessing as mp
from collections import defaultdict
import pymysql
import datajoint as dj
from datajoint.settings import default
from datajoint.utils import to_camel_case
from sshtunnel import SSHTunnelForwarder
from werkzeug.utils import secure_filename

from loris.database.attributes import custom_attributes_dict
from loris.utils import is_manuallookup
from loris.errors import LorisError


# defaults for application
defaults = {
    "textarea_startlength": 512,
    # UPLOAD EXTENSIONS
    "extensions": ['csv', 'npy', 'json', 'pkl', 'npz'],
    "attach_extensions": (
        ['csv', 'npy', 'json', 'pkl', 'npz'] + [
            'tiff', 'png', 'jpeg', 'mpg', 'hdf', 'hdf5', 'tar', 'zip',
            'txt', 'gif', 'svg', 'tif', 'bmp', 'doc', 'docx', 'rtf',
            'odf', 'ods', 'gnumeric', 'abw', 'xls', 'xlsx', 'ini',
            'plist', 'xml', 'yaml', 'yml', 'py', 'js', 'php', 'rb', 'sh',
            'tgz', 'txz', 'gz', 'bz2', 'jpe', 'jpg', 'pdf',
            'irradcal', 'cal'  # calibration files
        ]
    ),
    "additional_extensions": [],  # TODO
    # foreign key select field limit
    "fk_dropdown_limit": 200,
    "user_schema": "experimenters",
    "user_table": "Experimenter",
    "user_name": "experimenter",
    "group_schema": "experimenters",
    "group_table": "ExperimentalProject",
    "group_name": "experimental_project",
    "assignedgroup_schema": "experimenters",
    "assignedgroup_table": "AssignedExperimentalProject",
    "max_cpu": None,
    "init_database": True,
    "include_fly": False,
    "import_schema_module": {},  # schema_name : file_location
    "tables_skip_permission": ['`subjects`.`fly_stock`'],
    "filestores": {
        "attachstore": None,
        "datastore": None
    },
    "database.host": None,
    "database.port": None,
    "database.user": None,
    "database.password": None,
    "connection.charset": "utf8",
    "enable_python_native_blobs": True,
    "enable_python_pickle_blobs": True,
    "enable_automakers": True,
    "secret_key": "myprecious",
    "tmp_folder": "~/tmp",
    "skip_schemas": ["mysql", "sys", "performance_schema", "tutorial"],
    "user_schema": "experimenters",
    "user_table": "Experimenter",
    "user_name": "experimenter",
    "user_active": "active",
    "group_schema": "experimenters",
    "group_table": "ExperimentalProject",
    "group_name": "experimental_project",
    "assignedgroup_schema": "experimenters",
    "assignedgroup_table": "AssignedExperimentalProject",
    "standard_password": "fruitfly",
    "administrator_info": {
        "experimenter": "administrator",
        "first_name": "FIRSTNAME",
        "last_name": "LASTNAME",
        "email": "administrator@mail.com",
        "phone": "000-000-0000",
        "date_joined": "2020-01-01",
        "active": True
    },
    "administrators": ["administrator"],
    "wiki_folder": "~/loris/wiki",
    "autoscript_folder": "~/loris/autoscripts",
    "external_wiki": "#",
    "ssh_address": None,
    "ssh_username": "administrator",
    "ssh_pkey": "~/.ssh/id_rsa",
    "slack": [],
    "sql_privileges": {},
    # tables skipped to check for permission
}
AUTOSCRIPT_CONFIG = 'config.json'
USER_FILENAME = '._loris_config.json'
USER_CONFIG = os.path.join(os.path.expanduser('~'), USER_FILENAME)
GLOBAL_CONFIG = os.path.join(os.path.dirname(__file__), 'global_config.json')
EXPANDUSER_FIELDS = (
    'tmp_folder',
    'wiki_folder',
    'autoscript_folder',
    'ssh_pkey'
)


class Config(dict):

    _slack_tables = None

    @classmethod
    def load(cls, config_file):
        """load configuration class and perform necessary checks
        """

        if config_file is None:
            config_file = USER_CONFIG

            if not os.path.exists(config_file):
                config_file = 'loris_config.json'

                if not os.path.exists(config_file):
                    root_dir = os.path.dirname(os.path.dirname(__file__))
                    config_file = os.path.join(root_dir, 'config.json')

        if os.path.exists(config_file):
            with open(config_file, 'r') as f:
                config = json.load(f)
            with open(USER_CONFIG, 'w') as f:
                json.dump(config, f)
        elif os.path.exists(GLOBAL_CONFIG):
            with open(GLOBAL_CONFIG, 'r') as f:
                config = json.load(f)
        else:
            config = {}

        config = {**defaults, **config}
        config = cls(config)
        config['custom_attributes'] = custom_attributes_dict
        config['_empty'] = []  # list of files in tmp to delete on refresh
        config['_autopopulate'] = {}  # dictionary of subprocesses
        # delete permissions for datajoint
        config['delete_permission'] = config.datajoint_delete_permission
        config.perform_checks()
        config.datajoint_configuration()

        return config

    def datajoint_delete_permission(self, table):
        return self.user_has_permission(table, self['database.user'])

    def user_has_permission(self, table, user, skip_tables=None):
        """test if user is allowed to delete an entry or perform another action
        on a datajoint.Table
        """

        if user is None:
            user = self['connection'].get_user().split('@')[0]

        if user in self['administrators']:
            return True

        if table.database in self.groups_of_user(user):
            return True

        if skip_tables is None:
            skip_tables = self['tables_skip_permission']
            # skip_tables = []

        # always add table name
        skip_tables.append(table.full_table_name)

        if not table.connection.dependencies:
            table.connection.dependencies.load()

        ancestors = table.ancestors()

        if self.user_table.full_table_name in ancestors:
            if self['user_name'] in table.heading:
                user_only = table & {self['user_name']: user}
                return len(user_only) == len(table)
            else:
                for parent_name, parent_info in table.parents(foreign_key_info=True):
                    if parent_name in skip_tables:
                        continue

                    # get parent table
                    parent_table = self.get_table(parent_name)

                    # project only necessary keys
                    to_rename = {
                        ele: key
                        for key, ele in parent_info['attr_map'].items()
                    }
                    restricted_table = parent_table & table.proj(**to_rename)

                    if not self.user_has_permission(
                        restricted_table, user, skip_tables
                    ):
                        return False

        # checks if children have a parent table that is dependent on user table
        for child_name, child_info in table.children(foreign_key_info=True):
            if child_name in skip_tables:
                continue

            # get child table
            child_table = self.get_table(child_name)

            # restrict only with necessary keys
            restricted_table = child_table & table.proj(**child_info['attr_map'])

            if not self.user_has_permission(
                restricted_table, user, skip_tables
            ):
                return False

        return True

    @property
    def slacks(self):
        return self.get('slack', [])

    @property
    def slack_tables(self):
        if self._slack_tables is None:
            slack_tables = {}
            for slack in self.slacks:
                if slack.get('token', None) is not None:
                    slack_tables[slack['table']] = slack

            self._slack_tables = slack_tables

        return self._slack_tables

    def connect_ssh(self):
        """ssh tunneling

        see: https://sshtunnel.readthedocs.io/en/latest/
        """

        if self.get('server', None) is not None:
            return self['server']

        elif self.get('ssh_address', None) is not None:
            # inti kwargs for init
            remote_bind_address = (
                self['database.host'], self['database.port']
            )
            kwargs = {
                'remote_bind_address': remote_bind_address,
                'local_bind_address': remote_bind_address
            }

            # get signature
            signature = inspect.signature(SSHTunnelForwarder)

            for key, param in signature.parameters.items():
                if param.name in self:
                    kwargs[param.name] = self[param.name]

            kwargs.pop('ssh_address', None)
            print('parameters for ssh tunneling:')
            print(kwargs)

            address = self['ssh_address'].split(':')
            address[1] = int(address[1])
            address = tuple(address)
            print(address)

            # mute exceptions
            kwargs['mute_exceptions'] = True
            kwargs['raise_exception_if_any_forwarder_have_a_problem'] = False

            # initialize server
            server = SSHTunnelForwarder(
                address, **kwargs
            )
            self['server'] = server

            # start server
            server.start()

            return self['server']

    def disconnect_ssh(self):
        """ssh tunneling
        """

        if self.get('server', None) is not None:
            self['server'].stop()
            self['server'] = None

    def close(self):
        self['connection'].close()
        self.pop('connection', None)
        self.disconnect_ssh()

    def __getitem__(self, k):

        try:
            return super().__getitem__(k)
        except KeyError as e:
            if k == 'tables':
                self.refresh_tables()
                return self[k]
            elif k == 'schemata':
                self.refresh_schema()
                return self[k]
            elif k == 'dynamicforms':
                self[k] = {}
                return self[k]
            elif k == 'automaker_tables':
                self.refresh_automaker_tables()
                return self[k]
            elif k == 'settings_tables':
                self.refresh_settings_tables()
                return self[k]
            elif k == 'connection':
                return self.conn(reset=True)
            elif k == 'server':
                return self.connect_ssh()

            raise e

    def conn(self, *args, **kwargs):
        """connect to database with hostname, username, and password.
        """
        newargs = dict(zip(['host', 'user', 'password'], args))
        newargs.update(kwargs)
        already_connected = 'connection' in self
        # self.datajoint_configuration()
        self.connect_ssh()
        reconfigure_dj = False
        # database host
        if self['database.host'] is None and 'host' not in newargs and not already_connected:
            host = input(
                "What is the host address for your MySQL "
                "instance (defaults to `127.0.0.1`)? "
            )
            if not host:
                host = '127.0.0.1'
            self['database.host'] = host
            reconfigure_dj = True
        # database port
        if self['database.port'] is None and 'port' not in newargs and not already_connected:
            port = input(
                "What is the port for your MySQL "
                "instance (defaults to `3306`)? "
            )
            if not port:
                port = 3306
            else:
                port = int(port)
            self['database.port'] = port
            reconfigure_dj = True
        if self['database.user'] is None and 'user' not in newargs and not already_connected:
            user = input("Please enter your Loris/MySQL username: ")
            self['database.user'] = user
            reconfigure_dj = True
        if self['database.password'] is None and 'password' not in newargs and not already_connected:
            pw = getpass(prompt="Please enter Loris/MySQL password: ")
            self['database.password'] = pw
            reconfigure_dj = True

        if reconfigure_dj:
            self.datajoint_configuration()

        if self['database.host'] == 'mysql' and 'host' not in newargs:
            try:
                self['connection'] = dj.conn(**newargs)
            except pymysql.OperationalError:
                newargs.pop('host')
                self['connection'] = dj.conn('127.0.0.1', **newargs)
        else:
            self['connection'] = dj.conn(**newargs)

        # update backup context with schemata
        if newargs.get('refresh', False):
            self.refresh_schema()
        dj.config['backup_context'].update(self['schemata'])
        return self['connection']

    def datajoint_configuration(self):
        # --- managing external file stores for database --- #
        if 'stores' not in dj.config:
            dj.config['stores'] = {}

        for filestore_name, filestore in self['filestores'].items():
            if filestore is None:
                filestore = input(
                    f"What is the directory for your `{filestore_name}` "
                    f"(defaults to `~/loris/{filestore_name}`)? "
                )
                if not filestore:
                    filestore = f"~/loris/{filestore_name}"
                self['filestores'][filestore_name] = filestore
            filestore = os.path.expanduser(filestore)
            if not os.path.exists(filestore):
                os.makedirs(filestore)

            dj.config['stores'].update({
                filestore_name: {
                    'protocol': 'file',
                    'location': filestore
                }
            })

        # set datajoint variable in datajoint config
        for key, ele in default.items():
            if key in self:
                dj.config[key] = self[key]
            else:
                self[key] = ele

        dj.config['backup_context'] = custom_attributes_dict

    def perform_checks(self):
        """perform various checks (create directories if they don't exist)
        """

        for path in EXPANDUSER_FIELDS:
            if path in self:
                self[path] = os.path.expanduser(self[path])

        if sys.platform in ['win32', 'cygwin']:
            # enforce if on Windows
            self['tmp_folder'] = '/tmp'
        if not os.path.exists(self['tmp_folder']):
            os.makedirs(self['tmp_folder'])

        self['autoscript_folder'] = os.path.realpath(
            self['autoscript_folder'])
        if not os.path.exists(self['autoscript_folder']):
            os.makedirs(self['autoscript_folder'])

        if self['max_cpu'] is None:
            self['max_cpu'] = mp.cpu_count()

    def refresh_schema(self):
        """refresh container of schemas
        """
        schemata = {}

        # direct loading if possible
        # TODO (also in app init)
        if self['init_database']:
            from loris.schema import (
                experimenters, core
            )

            schemata['experimenters'] = experimenters
            schemata['core'] = core

            if self['include_fly']:
                from loris.schema import anatomy, subjects
                schemata['anatomy'] = anatomy  # move out
                schemata['subjects'] = subjects

        for schema, module_path in self["import_schema_module"]:
            # TODO test
            spec = importlib.util.spec_from_file_location(schema, module_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            schemata[schema] = module

        for schema in dj.list_schemas():
            if schema in self["skip_schemas"]:
                continue
            if schema in schemata:
                continue
            # TODO error messages
            schemata[schema] = dj.VirtualModule(
                schema, schema, connection=self['connection'],
                add_objects=custom_attributes_dict,
                create_tables=True
            )
            # make sure jobs table has been created
            schemata[schema].schema.jobs

        self['schemata'] = schemata

    def get_table_from_classname(self, class_name):
        """get table from schemata using table class name format
        """

        table_info = class_name.split('.')
        schema, table_classes = table_info[0], table_info[1:]

        schema_module = self['schemata'].get(schema, None)

        if schema_module is None:
            raise LorisError(
                f'schema {schema} not in database; refresh database'
            )

        table = schema_module

        for table_ in table_classes:

            try:
                table = getattr(table, table_)
            except AttributeError:
                raise LorisError(
                    f'table {table_} not in schema {schema}; '
                    'refresh database'
                )

        return table

    def get_table(self, full_table_name, as_string=False):
        """get table from schemata using full_table_name
        """

        schema, table_name = full_table_name.replace('`', '').split('.')

        schema_module = self['schemata'].get(schema, None)

        if schema_module is None:
            raise LorisError(
                f'schema {schema} not in database; refresh database'
            )

        table_name = table_name.strip('_#')
        table_name_list = table_name.split('__')
        if as_string:
            output = schema
        else:
            output = schema_module

        for tname in table_name_list:
            tname = to_camel_case(tname)
            try:
                if as_string:
                    output = '.'.join([output, tname])
                else:
                    output = getattr(output, tname)
            except AttributeError:
                raise LorisError(
                    f'table {table_name} not in schema {schema}; '
                    'refresh database'
                )

        return output

    def refresh_tables(self):
        """refresh container of tables
        """

        tables = {}

        for schema, module in self['schemata'].items():
            # skip mysql schema etc
            if schema in self["skip_schemas"]:
                continue
            for key, ele in module.__dict__.items():
                if key.split('.')[0] in self['schemata']:
                    continue

                if inspect.isclass(ele) and issubclass(ele, dj.user_tables.UserTable):
                    if is_manuallookup(ele) or issubclass(ele, dj.Settingstable):
                        continue
                    tables[f'{schema}.{key}'] = ele

                    # get part tables
                    for part_name, part_table in ele.__dict__.items():
                        if inspect.isclass(part_table):
                            if issubclass(part_table, dj.Part):
                                tables[f'{schema}.{key}.{part_name}'] = \
                                    part_table

        self['tables'] = tables

        return tables

    @property
    def user_table(self):
        """return the user table
        """

        return getattr(
            self['schemata'][self['user_schema']],
            self['user_table'])

    @property
    def users(self):
        """get a list of all users
        """
        users = list(
            self.user_table.proj(self['user_name']).fetch()[self['user_name']]
        )

        if not users:
            self.create_administrator()

        return users

    def create_administrator(self):
        # insert administrator if not users exist and create
        self.user_table.insert1(self['administrator_info'])

        # use standard password
        password = self['standard_password']
        # establish connection
        conn = self['connection']
        connection = '%'
        username = self['administrator_info'][self['user_name']]

        # for safety flush all privileges
        conn.query("FLUSH PRIVILEGES;")

        conn.query(
            "DROP USER IF EXISTS %s@%s;",
            (username, connection)
        )
        conn.query(
            "CREATE USER %s@%s IDENTIFIED BY %s;",
            (username, connection, password)
        )

        # create user-specific schema
        schema = dj.Schema(username)

        privileges = {
            '*.*': "ALL PRIVILEGES",
        }

        for dbtable, privilege in privileges.items():
            privilege = (f"GRANT {privilege} ON {dbtable} to %s@%s;")
            conn.query(privilege, (username, connection))

        conn.query("FLUSH PRIVILEGES;")

        return schema

    @property
    def user_tables(self):
        """return a list of user tables
        """
        return list(set(self.users) & set(self['schemata']))

    @property
    def group_table(self):
        """return the group table
        """

        return getattr(
            self['schemata'][self['group_schema']],
            self['group_table'])

    @property
    def groups(self):
        """get a list of all groups
        """
        groups = list(
            self.group_table.proj(
                self['group_name']
            ).fetch()[self['group_name']]
        )

        return groups

    @property
    def group_tables(self):
        """return a list of group tables
        """
        return list(set(self.groups) & set(self['schemata']))

    def create_group_schemas(self):
        """create all missing schemas
        """

        for group in self.groups:
            dj.Schema(group, connection=self['connection'])

    @property
    def assigned_table(self):
        """assigned table (matching groups and users)
        """

        return getattr(
            self['schemata'][self['assignedgroup_schema']],
            self['assignedgroup_table'])

    def groups_of_user(self, user):
        """groups user belongs to (includes user name)
        """

        groups = [user]

        table = self.assigned_table & {
            self['user_name'] : user
        }
        groups.extend(
            list(table.proj(self['group_name']).fetch()[self['group_name']])
        )

        return groups

    def schemas_of_user(self, user):
        """schemas user belongs to (should be the same as groups_of_user except
        when administrator), if each group has an associated schema.
        """

        if user in self['administrators']:
            return list(self['schemata'])

        groups = self.groups_of_user(user)
        return list(set(groups) & set(self['schemata']))

    def user_in_group(self, user, group):
        """is user in group
        """

        table = self.assigned_table & {
            self['user_name'] : user,
            self['group_name'] : group
        }

        if len(table) == 1:
            return True
        elif len(table) == 0:
            return False
        else:
            raise LorisError(
                "assined user group table should only have "
                "singular entries for given user and group."
            )

    def refresh_permissions(self):
        """refresh permissions of users
        """

        try:
            for user in self.users:
                for schema in self.schemas_of_user(user):
                    conn = self['connection']
                    conn.query("FLUSH PRIVILEGES;")
                    conn.query(
                        f"GRANT ALL PRIVILEGES ON {schema}.* to %s@%s;",
                        (user, '%')
                    )
                    conn.query("FLUSH PRIVILEGES;")
        except Exception as e:
            print(f"Did not refresh permissions! Due to \n {e}")

    def refresh_settings_tables(self):
        """refresh container of settings table
        """

        tables = {}

        for table_name, table in self['tables'].items():
            if issubclass(table, dj.Settingstable):
                tables[table_name] = table

        self['settings_tables'] = tables

        return tables

    def refresh_automaker_tables(self):
        """refresh container of settings table
        """

        tables = {}

        for table_name, table in self['tables'].items():
            if issubclass(table, (dj.AutoImported, dj.AutoComputed)):
                tables[table_name] = table

        self['automaker_tables'] = tables

        return tables

    def tables_to_list(self):
        """convert tables container to list for app header
        """

        tables = self['tables']
        manualtables_dict = defaultdict(list)
        autotables_list = []

        for table_name, table in tables.items():
            if issubclass(table, dj.Manual):
                auto_class = False
                # ignore ManualLookup subclasses
                if (
                    is_manuallookup(table)
                    or (table.full_table_name
                        == self.user_table.full_table_name)
                    or (table.full_table_name
                        == self.group_table.full_table_name)
                    or (table.full_table_name
                        == self.assigned_table.full_table_name)
                ):
                    continue
            elif issubclass(table, (dj.AutoImported, dj.AutoComputed)):
                auto_class = True
            else:
                continue

            table_list = [table_name] + table_name.split('.') + [None]

            if auto_class:
                autotables_list.append(table_list)
            else:
                manualtables_dict[table_list[1]].append(table_list[2])

        return manualtables_dict, autotables_list

    def refresh_dependencies(self):
        """refresh dependencies of database connection
        """

        self['connection'].dependencies.load()

    def empty_tmp_folder(self):
        """empty temporary folder
        """

        for filename in os.listdir(self['tmp_folder']):
            if not any([filename.startswith(ele) for ele in self['_empty']]):
                continue
            file_path = os.path.join(self['tmp_folder'], filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print(f'Failed to delete {file_path}. Reason: {e}')

        self['_empty'] = []

    def refresh(self):
        """refresh all containers and empty temporary folder
        """

        self.pop('dynamicforms', None)
        self.pop('autoscriptforms', None)
        self.empty_tmp_folder()
        self.refresh_dependencies()
        self.refresh_schema()
        self.refresh_tables()
        self.refresh_settings_tables()
        self.refresh_automaker_tables()
        self.refresh_permissions()

    def get_dynamicform(
        self, table_name, table_class, dynamic_class, **kwargs
    ):
        """get the dynamic form and wtf form for application
        """

        name = dynamic_class.__name__

        if name not in self['dynamicforms']:
            self['dynamicforms'][name] = {}

        if table_name not in self['dynamicforms'][name]:
            dynamicform = dynamic_class(table_class)
            form = dynamicform.formclass(**kwargs)
            self['dynamicforms'][name][table_name] = dynamicform
        else:
            # update foreign keys
            dynamicform = self['dynamicforms'][name][table_name]
            form = dynamicform.formclass(**kwargs)
            dynamicform.update_fields(form)

        return dynamicform, form

    def get_autoscriptforms(
        self, autoscript_filepath, table_name, form_creator, **kwargs
    ):
        """get autoscript form and process_dict and buttons params
        """
        foldername = os.path.basename(autoscript_filepath)
        filepath = os.path.join(
            autoscript_filepath, AUTOSCRIPT_CONFIG
        )

        with open(filepath, 'r') as f:
            try:
                config = json.load(f)
            except Exception as e:
                raise LorisError(
                    f"Json config file for autoscript"
                    f" '{foldername}' badly "
                    f"formatted: {e}"
                )

        if 'buttons' not in config:
            raise LorisError(
                "In configuration file of autoscript "
                f"{foldername}, no 'button' key "
                "was provided."
            )

        buttons = config['buttons']
        config_forms = config.get('forms', {})

        if not isinstance(buttons, dict):
            raise LorisError(
                f'In configuration file of autoscript '
                f'"{foldername}", '
                '"scripts" keyword is incorrectly '
                'formatted.'
            )

        for key, button in buttons.items():
            base_message = (
                f'In configuration file of autoscript '
                f'"{foldername}", '
                '"buttons" keyword is incorrectly '
                f'formatted for button "{key}":'
            )
            message = (
                f'{base_message} must be dictionary with '
                '"script" key and "validate", "insert", "configattr", '
                '"outputfile", and "outputattr" optionally defined.'
            )
            if not isinstance(button, dict):
                raise LorisError(message)
            elif not all([
                isinstance(button.get('script', None), str),
                isinstance(button.get('validate', []), list),
                isinstance(button.get('insert', False), bool),
                isinstance(button.get('configattr', ''), str),
                isinstance(button.get('outputfile', ''), (str, list)),
                isinstance(button.get('outputattr', ''), (str, list)),
            ]):
                raise LorisError(message)

            button['script'] = secure_filename(button['script'])
            if not os.path.exists(
                os.path.join(autoscript_filepath, button['script'])
            ):
                raise LorisError(
                    f'{base_message} script "{button["script"]}" '
                    'does not exist in autoscript folder.'
                )

            if (
                'outputfile' in button
                and isinstance(button['outputfile'], str)
            ):
                if not isinstance(button['outputattr'], str):
                    raise LorisError(message)
                # put into list format
                button['outputfile'] = [secure_filename(button['outputfile'])]
                button['outputattr'] = [button['outputattr']]
            elif 'outputfile' in button:
                if not (
                    isinstance(button['outputattr'], list)
                    and len(button['outputfile']) == len(button['outputattr'])
                ):
                    raise LorisError(message)

                button['outputfile'] = [
                    secure_filename(ifilepath)
                    for ifilepath in button['outputfile']
                ]

        if not isinstance(config_forms, dict):
            raise LorisError(
                f'In configuration file of autoscript '
                f'"{os.path.basename(autoscript_filepath)}", '
                '"forms" keyword is incorrectly '
                'formatted.'
            )

        forms = {}
        post_process_dict = {}
        for key, value in config_forms.items():
            form, post_process = form_creator(
                value, autoscript_filepath, **kwargs
            )
            forms[key] = form
            post_process_dict[key] = post_process

        return forms, post_process_dict, buttons



"""run populate
"""

import argparse
import sys
import os
import json


if __name__ == '__main__':
    print('starting to populate')
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--schema", help="name of schema", type=str)
    parser.add_argument(
        "--table", help="name of table", type=str)
    parser.add_argument(
        "--kwargs", help="keyword arguments to pass to populate", type=str
    )
    parser.add_argument(
        "--user", help="username for sql database", type=str
    )
    parser.add_argument(
        "--password", help="password for sql database", type=str
    )

    args = parser.parse_args()
    # TODO connect as user?

    kwargs = json.loads(args.kwargs)

    # add loris to path if not installed
    try:
        from loris import config, conn
    except (ModuleNotFoundError, ImportError):
        filepath = __file__
        for i in range(4):
            filepath = os.path.dirname(filepath)
        print(filepath)
        sys.path.append(filepath)
        from loris import config, conn

    conn(user=args.user, password=args.password)

    table_class = getattr(
        config['schemata'][args.schema],
        args.table
    )

    restriction = kwargs.pop('restriction')
    settings_name = kwargs.pop('settings_name')
    print("settings_name:")
    print(settings_name)
    print("restriction:")
    print(restriction)

    table_class.populate(settings_name, restriction, **kwargs)



"""processor class
"""

from collections import OrderedDict

import markdown

from loris.app.wiki.utils import wikilink


class Processor(object):
    """
        The processor handles the processing of file content into
        metadata and markdown and takes care of the rendering.

        It also offers some helper methods that can be used for various
        cases.
    """

    preprocessors = []
    postprocessors = [wikilink]

    def __init__(self, text):
        """
            Initialization of the processor.

            :param str text: the text to process
        """
        self.md = markdown.Markdown(extensions=[
            'codehilite',
            'fenced_code',
            'meta',
            'tables'
        ])
        self.input = text
        self.markdown = None
        self.meta_raw = None

        self.pre = None
        self.html = None
        self.final = None
        self.meta = None

    def process_pre(self):
        """
            Content preprocessor.
        """
        current = self.input
        for processor in self.preprocessors:
            current = processor(current)
        self.pre = current

    def process_markdown(self):
        """
            Convert to HTML.
        """
        self.html = self.md.convert(self.pre)

    def split_raw(self):
        """
            Split text into raw meta and content.
        """
        self.meta_raw, self.markdown = self.pre.split('\n\n', 1)

    def process_meta(self):
        """
            Get metadata.

            .. warning:: Can only be called after :meth:`html` was
                called.
        """
        # the markdown meta plugin does not retain the order of the
        # entries, so we have to loop over the meta values a second
        # time to put them into a dictionary in the correct order
        self.meta = OrderedDict()
        for line in self.meta_raw.split('\n'):
            key = line.split(':', 1)[0]
            # markdown metadata always returns a list of lines, we will
            # reverse that here
            self.meta[key.lower()] = \
                '\n'.join(self.md.Meta[key.lower()])

    def process_post(self):
        """
            Content postprocessor.
        """
        current = self.html
        for processor in self.postprocessors:
            current = processor(current)
        self.final = current

    def process(self):
        """
            Runs the full suite of processing on the given text, all
            pre and post processing, markdown rendering and meta data
            handling.
        """
        self.process_pre()
        self.process_markdown()
        self.split_raw()
        self.process_meta()
        self.process_post()

        return self.final, self.markdown, self.meta
